###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Module for creating, describing and managing production requests objects."""
import copy
import re

from DIRAC import gLogger, S_OK

from DIRAC.Core.Utilities.DAG import DAG
from DIRAC.Core.Workflow.Workflow import fromXMLString
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations

from LHCbDIRAC.Interfaces.API.DiracProduction import DiracProduction
from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
from LHCbDIRAC.ProductionManagementSystem.Client.Production import Production
from LHCbDIRAC.ProductionManagementSystem.Client.ProductionRequestClient import ProductionRequestClient
from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient


class ProductionRequest:
    """Production request class - objects are usually created starting from a production request"""

    def __init__(self, bkkClientIn=None, diracProdIn=None):
        """c'tor.

        Some variables are defined here. A production request is made of:
        stepsList, productionsTypes, and various parameters of those
        productions
        """

        if bkkClientIn is None:
            self.bkkClient = BookkeepingClient()
        else:
            self.bkkClient = bkkClientIn

        if diracProdIn is None:
            self.diracProduction = DiracProduction()
        else:
            self.diracProduction = diracProdIn

        self.rpcProductionRequest = ProductionRequestClient()
        self.tc = TransformationClient()

        self.logger = gLogger.getSubLogger("ProductionRequest")

        self.opsH = Operations()

        # parameters of the request
        self.requestID = 0
        self.parentRequestID = 0
        self.appendName = "1"
        self.outConfigName = ""
        self.prodsToLaunch = []  # productions to launch
        self.stepsListDict = []  # list of dict of steps
        self.stepsInProds = []  # a list of lists
        self.mcTesting = True
        # parameters of the input data
        self.processingPass = ""
        self.dataTakingConditions = ""
        self.eventType = ""
        self.bkFileType = []
        self.dqFlag = ""
        self.extendedDQOK = ""
        self.smog2State = ""
        self.bkSampleSeed = None
        self.bkSampleMax = None
        self.startRun = 1
        self.endRun = 2
        self.runsList = ""
        self.configName = "test"
        self.configVersion = "certification"
        # In DAG-like workflows, "applyOptionalCorrections" can horribly break things
        # so provide an option to turn it off.
        self.disableOptionalCorrections = False
        # parameters of the first production
        self.publishFlag = True
        self.testFlag = False
        self.derivedProduction = 0
        self.previousProdID = 0  # optional prod from which to start
        self.fullListOfOutputFileTypes = []

        # parameters that are the same for each productions
        self.prodGroup = ""  # This ends up being 'ProcessingType', workflow parameter, and it is used for accounting
        self.visibility = ""  # For BKQuery
        self.fractionToProcess = 0
        self.minFilesToProcess = 0
        self.modulesList = None  # Usually:
        # ['GaudiApplication', 'AnalyseXMLSummary',
        # 'ErrorLogging', 'BookkeepingReport', 'StepAccounting' ]
        self.extraModulesList = None

        # parameters of each production (the length of each list has to be the same as the number of productions
        self.events = []
        self.stepsList = []
        self.extraOptions = {}
        self.eventTimeouts = {}
        self.prodsTypeList = []
        self.bkQueries = []  # list of bk queries
        self.removeInputsFlags = []
        self.priorities = []
        self.cpus = []
        self.inputs = []  # list of lists
        self.outputModes = []
        self.targets = []
        self.outputFileMasks = []
        self.outputFileSteps = []
        self.groupSizes = []
        self.plugins = []
        self.inputDataPolicies = []
        self.previousProds = [None]  # list of productions from which to take the inputs (the first is always None)
        self.multicore = []  # list of flags to override the multi core flags of the steps
        # list of tuples with min and max number of processors allowed for the jobs ((0,0) = unlimited)
        self.processors = []

        self.outputSEs = []  # a list of StorageElements
        self.specialOutputSEs = []  # a list of dictionaries - might be empty
        self.outputSEsPerFileType = []  # a list of dictionaries - filled later
        self.ancestorDepths = []
        self.compDict = {"HIGH": "Compression-LZMA-4", "LOW": "Compression-ZLIB-1"}
        self.compressionLvl = [""]  # List: one compression level each step
        self.appConfig = "$APPCONFIGOPTS/Persistency/"  # Default location of the compression level configuration files
        #
        # These lists define the visibility of the output files produced by each step. For MC productions, the visibility
        # is tied to compression level. VIsible files are compressed at the highest level
        #
        self.outputVisFlag = []  # List of dictionary with default visibility flag of the output files per single step
        self.specialOutputVisFlag = []  # List of dictionaries with special visibility flag for given file type

    #############################################################################

    def resolveSteps(self):
        """Given a list of steps in strings, some of which might be missing,
        resolve it into a list of dictionary of steps (self.stepsListDict)"""
        outputVisFlag = {
            k: v for el in self.outputVisFlag for k, v in el.items()
        }  # Transform the list of dictionaries in a dictionary
        specialOutputVisFlag = {k: v for el in self.specialOutputVisFlag for k, v in el.items()}
        count = 0  # Needed to add correctly the optionFiles to the list of dictonaries of steps
        for stepID in self.stepsList:
            stepDict = self.bkkClient.getAvailableSteps({"StepId": stepID})
            if not stepDict["OK"]:
                raise ValueError(stepDict["Message"])
            else:
                stepDict = stepDict["Value"]

            stepsListDictItem = {}

            s_in = self.bkkClient.getStepInputFiles(stepID)
            if not s_in["OK"]:
                raise ValueError(s_in["Message"])
            stepsListDictItem["fileTypesIn"] = [fileType[0].strip() for fileType in s_in["Value"]["Records"]]

            s_out = self.bkkClient.getStepOutputFiles(stepID)
            if not s_out["OK"]:
                raise ValueError(s_out["Message"])
            fileTypesList = [fileType[0].strip() for fileType in s_out["Value"]["Records"]]
            self.fullListOfOutputFileTypes = self.fullListOfOutputFileTypes + fileTypesList
            stepsListDictItem["fileTypesOut"] = fileTypesList

            for parameter, value in zip(stepDict["ParameterNames"], stepDict["Records"][0]):
                if parameter.lower() in ["conddb", "dddb", "dqtag"] and value:
                    if value.lower() == "frompreviousstep":
                        value = self.stepsListDict[-1][parameter]

                # TODO: This should be moved to LbProdRun
                if (
                    parameter == "OptionFiles"
                    and "MDF" not in stepsListDictItem["fileTypesOut"]
                    and not value.startswith("{")
                ):
                    # Modifying the OptionFiles (for setting the compression level)
                    # certain MC produce MDF, which shouldn't be compressed
                    #
                    # If the prod manager sets a compression level for a particular step, either we append the option file
                    # or we overwrite the existing one inherited with the step
                    #
                    if len(self.compressionLvl) > count and self.compressionLvl[count] != "":
                        persist = re.compile("Compression-[A-Z]{4}-[1-9]")
                        # self.compressionLvl[count] = self.appConfig + self.compressionLvl[count] + '.py'
                        self.compressionLvl[count] = (
                            self.appConfig + self.compDict[self.compressionLvl[count].upper()] + ".py"
                        )
                        if not persist.search(value):
                            if value == "":
                                value = self.compressionLvl[count]
                            else:
                                value = ";".join((value, self.compressionLvl[count]))
                        else:
                            value = persist.sub(persist.search(self.compressionLvl[count]).group(), value)
                    #
                    # If instead the prod manager doesn't declare a compression level, e.g. for intermediate steps,
                    # we check if there is one in the options and in case we delete it. This leaves the default zip level
                    # defined inside Gaudi
                    #
                    elif len(self.compressionLvl) > count and self.compressionLvl[count] == "":
                        persist = re.compile(r"\$\w+/Persistency/Compression-[A-Z]{4}-[1-9].py;?")
                        if persist.search(value):
                            value = persist.sub("", value)

                if parameter == "SystemConfig" and value is not None and re.search("slc5", value):
                    # TODO: Remove this from the DB so the hack can be removed
                    p = re.compile(r"\$\w+/Persistency/Compression-[A-Z]{4}-[1-9].py;?")
                    if p.search(stepsListDictItem["OptionFiles"]):
                        stepsListDictItem["OptionFiles"] = p.sub("", stepsListDictItem["OptionFiles"])

                stepsListDictItem[parameter] = value  # Fixing what decided

            if stepsListDictItem["StepId"] in self.extraOptions:
                stepsListDictItem["ExtraOptions"] = self.extraOptions[stepsListDictItem["StepId"]]
            else:
                stepsListDictItem["ExtraOptions"] = ""

            if stepsListDictItem["StepId"] in self.eventTimeouts:
                stepsListDictItem["EventTimeout"] = self.eventTimeouts[stepsListDictItem["StepId"]]
            else:
                stepsListDictItem["EventTimeout"] = ""

            stepsListDictItem["prodStepID"] = str(stepID) + str(stepsListDictItem["fileTypesIn"])

            if "isMulticore" not in stepsListDictItem:
                stepsListDictItem["isMulticore"] = "N"

            if "SystemConfig" not in stepsListDictItem:
                stepsListDictItem["SystemConfig"] = ""

            if "mcTCK" not in stepsListDictItem:
                stepsListDictItem["mcTCK"] = ""

            # Add visibility info during step resolution
            if "visibilityFlag" not in stepsListDictItem:
                outputVisList = list(
                    {"Visible": outputVisFlag[str(count + 1)], "FileType": ftype}
                    for ftype in stepsListDictItem["fileTypesOut"]
                )
                if str(count + 1) in specialOutputVisFlag:
                    for it in outputVisList:
                        if it["FileType"] in specialOutputVisFlag[str(count + 1)]:
                            it["Visible"] = specialOutputVisFlag[str(count + 1)][it["FileType"]]

                stepsListDictItem["visibilityFlag"] = outputVisList

            self.stepsListDict.append(stepsListDictItem)
            count += 1

    #############################################################################

    def buildAndLaunchRequest(self):
        """uses _applyOptionalCorrections, _getProdsDescriptionDict,
        _buildProduction, and DiracProduction.launchProduction."""

        if not self.stepsListDict:
            self.resolveSteps()

        if not self.disableOptionalCorrections:
            self._applyOptionalCorrections()

        self._determineOutputSEs()

        prodsDict = self._getProdsDescriptionDict()

        stepsListDict = list(self.stepsListDict)

        fromProd = self.previousProdID
        prodsLaunched = []

        self.logger.debug(prodsDict)
        # now we build and launch each productions
        for prodIndex, prodDict in prodsDict.items():
            if self.prodsToLaunch:
                if prodIndex not in self.prodsToLaunch:
                    continue

            # build the list of steps in a production
            stepsInProd = []
            for stepID in prodDict["stepsInProd-ProdName"]:
                for step in stepsListDict:
                    if step["prodStepID"] == stepID:
                        stepsInProd.append(stepsListDict.pop(stepsListDict.index(step)))
            # NOT READY (alternative to previous 5 lines)
            # build the DAG of steps in a production
            # stepsInProdDAG = self._getStepsInProdDAG( prodDict, stepsListDict )
            # Here, for today it is just convert to a list
            # TODO: fix this in order to properly use DAGs (now it's only sequential)
            # FIXME: using getIndexNodes we can't assure the order is respected
            # stepsInProd = stepsInProdDAG.getIndexNodes()

            if prodDict["previousProd"] is not None:
                fromProd = prodsLaunched[prodDict["previousProd"] - 1]
                self.previousProdID = fromProd

            prod = self._buildProduction(
                prodType=prodDict["productionType"],
                stepsInProd=stepsInProd,
                outputSE=prodDict["outputSE"],
                priority=prodDict["priority"],
                cpu=prodDict["cpu"],
                inputDataList=prodDict["input"],
                outputMode=prodDict["outputMode"],
                inputDataPolicy=prodDict["inputDataPolicy"],
                outputFileMask=prodDict["outputFileMask"],
                outputFileStep=prodDict["outputFileStep"],
                target=prodDict["target"],
                removeInputData=prodDict["removeInputsFlag"],
                groupSize=prodDict["groupSize"],
                bkQuery=prodDict["bkQuery"],
                plugin=prodDict["plugin"],
                previousProdID=fromProd,
                derivedProdID=prodDict["derivedProduction"],
                transformationFamily=prodDict["transformationFamily"],
                events=prodDict["events"],
                multicore=prodDict["multicore"],
                processors=prodDict["processors"],
                ancestorDepth=prodDict["ancestorDepth"],
            )

            # if the production is a simulation production type, submit it to the automated testing
            if (
                prodDict["productionType"]
                in self.opsH.getValue("Transformations/ExtendableTransfTypes", ["MCSimulation"])
                and self.mcTesting
            ):
                prodID = self._mcSpecialCase(prod, prodDict)

            else:
                res = self.diracProduction.launchProduction(
                    prod=prod,
                    publishFlag=self.publishFlag,
                    testFlag=self.testFlag,
                    requestID=self.requestID,
                    tracking=prodDict["tracking"],
                )
                if not res["OK"]:
                    raise RuntimeError(res["Message"])

                prodID = res["Value"]

            prodsLaunched.append(prodID)

            if self.publishFlag:
                self.logger.notice(
                    f"For request {self.requestID}, submitted Production",
                    f"{prodIndex}, of type {prodDict['productionType']}, ID = {prodID}",
                )
        return S_OK(prodsLaunched)

    #############################################################################

    def _getStepsInProdDAG(self, prodDict, stepsListDict, stepsOrder="sequential"):
        """Builds the DAG of steps in a production.

        :params dict prodDict: dictionary representing one production
        :params list stepsListDict: list of steps (which are dictionaries) that should be in the production

        :returns: stepsInProd (DAG)
        """
        stepsInProd = DAG()

        inserted = None
        for stepID in prodDict["stepsInProd-ProdName"]:
            for step in stepsListDict:
                if step["prodStepID"] == stepID:
                    ind = stepsListDict.index(step)
                    step = stepsListDict.pop(ind)
                    stepsInProd.addNode(step)
                    if inserted and stepsOrder == "sequential":
                        stepsInProd.addEdge(inserted, step)
                    inserted = step

        return stepsInProd

    def _mcSpecialCase(self, prod, prodDict):
        """Treating the MC special case for putting MC productions in status
        "Testing"."""

        # save the original xml before it is edited for testing
        prod._lastParameters()  # pylint: disable=protected-access

        # launchProduction adds extra parameters, as we 'hot swap' the xml, we
        # need to get these parameters for the un-edited version
        originalProcessingType = prod.prodGroup
        originalPriority = prod.priority

        prodXML = prod.LHCbJob.workflow.toXML()

        prodID = self._modifyAndLaunchMCXML(prod, prodDict)

        # load a production from the original xml to save the priority and processing type
        workflowToSave = fromXMLString(prodXML)
        prod.LHCbJob.workflow = workflowToSave
        prod.setParameter("ProcessingType", "JDL", str(originalProcessingType), "ProductionGroupOrType")
        prod.setParameter("Priority", "JDL", str(originalPriority), "Job Priority")

        # original xml to save
        descriptionToStore = prod.LHCbJob.workflow.toXML()

        # saving the original xml in the StoredJobDescription table.
        res = self.tc.addStoredJobDescription(prodID, descriptionToStore)
        if not res["OK"]:
            self.logger.error("Error calling addStoredJobDescription", res["Message"])
            self.logger.info("Cleaning created production and exiting")
            self.diracProduction.production(res["Value"], "cleaning")
            raise RuntimeError(res["Message"])

        return prodID

    def _modifyAndLaunchMCXML(self, prod, prodDict):
        """Apply modifications to the workflow XML for MC testing case.

        :param Production prod: Production object
        :param dict prodDict: dictionary with production info

        :returns: res['OK'] or res['ERROR']
        """
        # set the destination and number of events for testing
        destination = self.opsH.getValue("Productions/MCTesting/MCTestingDestination", "DIRAC.Test.ch")
        numberOfEvents = self.opsH.getValue("Productions/MCTesting/numberOfEvents", "500")
        extendBy = self.opsH.getValue("Productions/MCTesting/extendBy", 20)

        prod.setJobParameters({"Destination": destination})
        prod.LHCbJob.workflow.removeParameter("BannedSites")
        prod.setParameter("numberOfEvents", "string", str(numberOfEvents), "Number of events to test")

        # add '1' to the stepMask and add GAUSSHIST to the fileMask
        fileTypesOutLastStep = []
        listOutputLastStep = prod.LHCbJob.workflow.step_instances[-2].findParameter("listoutput").getValue()
        for ols in listOutputLastStep:
            fileTypesOutLastStep.append(ols["outputDataType"])
        newFileMask = ["GAUSSHIST"] + [ftOut.upper() for ftOut in fileTypesOutLastStep]
        stepMaskParameter = prod.LHCbJob.workflow.findParameter("outputDataStep")
        if stepMaskParameter:
            stepMask = stepMaskParameter.getValue().replace(" ", "").split(";")
            newOutputFileStep = ";".join(sorted(list({"1"}.union(set(stepMask)))))
        else:
            newOutputFileStep = "1"
        prod.setFileMask(newFileMask, newOutputFileStep)

        # find the file types out already built, append GAUSSHIST and set the new listoutput
        fileTypesOut = []
        listOutput = prod.LHCbJob.workflow.step_instances[0].findParameter("listoutput").getValue()
        for ls in listOutput:
            fileTypesOut.append(ls["outputDataType"])
        fileTypesOut.append("GAUSSHIST")
        outputFilesList = prod._constructOutputFilesList(fileTypesOut)  # pylint: disable=protected-access
        prod.LHCbJob.workflow.step_instances[0].setValue("listoutput", outputFilesList)

        # increase the priority to 10
        prod.priority = 10

        # launch the test production
        res = self.diracProduction.launchProduction(
            prod=prod,
            publishFlag=self.publishFlag,
            testFlag=self.testFlag,
            requestID=self.requestID,
            extend=extendBy,
            tracking=prodDict["tracking"],
            MCsimflag=True,
        )
        if not res["OK"]:
            self.logger.error("Error launching production", res["Message"])
            raise RuntimeError(res["Message"])

        return res["Value"]

    def _determineOutputSEs(self):
        """Fill outputSEsPerFileType based on outputSEs, fullListOfOutputFileTypes
        and specialOutputSEs."""
        for outputSE, specialOutputSEs in zip(self.outputSEs, self.specialOutputSEs):
            outputSEDict = {}
            if not self.fullListOfOutputFileTypes:
                raise ValueError("No steps defined")
            outputSEDict = {fType: outputSE for fType in self.fullListOfOutputFileTypes}
            if specialOutputSEs:
                outputSEDict.update(specialOutputSEs)
            self.outputSEsPerFileType.append(outputSEDict)

    def _applyOptionalCorrections(self):
        """if needed, calls _splitIntoProductionSteps.

        It also applies other changes
        """
        if len(self.bkQueries) != len(self.prodsTypeList):
            self.bkQueries += ["fromPreviousProd"] * (len(self.prodsTypeList) - len(self.bkQueries))

        if len(self.previousProds) != len(self.prodsTypeList):
            self.previousProds += range(1, len(self.prodsTypeList))

        if len(self.events) != len(self.prodsTypeList):
            self.events += ["-1"] * (len(self.prodsTypeList) - len(self.events))

        if not self.removeInputsFlags:
            removeInputsFlags = []
            for prodType in self.prodsTypeList:
                if prodType.lower() == "merge":
                    removeInputsFlags.append(True)
                else:
                    removeInputsFlags.append(False)
            self.removeInputsFlags = removeInputsFlags

        if not self.outputFileMasks:
            self.outputFileMasks = [""] * len(self.prodsTypeList)

        if not self.outputFileSteps:
            self.outputFileSteps = [""] * len(self.prodsTypeList)

        if not self.inputs:
            self.inputs = [[]] * len(self.prodsTypeList)

        if not self.outputModes:
            self.outputModes = ["Any"] * len(self.prodsTypeList)

        if not self.targets:
            self.targets = [""] * len(self.prodsTypeList)

        if not self.inputDataPolicies:
            self.inputDataPolicies = ["download"] * len(self.prodsTypeList)

        if not self.multicore:
            self.multicore = ["True"] * len(self.prodsTypeList)

        if not self.processors:
            self.processors = [(0, 0)] * len(self.prodsTypeList)

        if not self.specialOutputSEs:
            self.specialOutputSEs = [{}] * len(self.prodsTypeList)

        if not self.ancestorDepths:
            self.ancestorDepths = [0] * len(self.prodsTypeList)

        # Checking if we need to split the merging step into many productions
        if "merge" in [pt.lower() for pt in self.prodsTypeList]:
            i = 0
            indexes = []
            for pt in self.prodsTypeList:
                if pt.lower() == "merge":
                    indexes.append(i)
                i += 1

            for index in indexes:
                # In this case and only in this case I have to split the merging in many productions
                plugin = self.plugins[index]
                outputSE = self.outputSEs[index]
                specialOutputSE = self.specialOutputSEs[index]
                priority = self.priorities[index]
                cpu = self.cpus[index]
                bkQuery = self.bkQueries[index]
                groupSize = self.groupSizes[index]
                preProd = self.previousProds[index]
                removeInputsFlag = self.removeInputsFlags[index]
                outputFileMask = self.outputFileMasks[index]
                outputFileStep = self.outputFileSteps[index]
                inputs = self.inputs[index]
                idp = self.inputDataPolicies[index]
                stepID = self.stepsList[index]
                events = self.events[index]
                targets = self.targets[index]
                multicore = self.multicore[index]
                processors = self.processors[index]
                outputMode = self.outputModes[index]
                ancestorDepth = self.ancestorDepths[index]
                if (
                    plugin.lower() not in ["approcessingbyfiletypesize", "byrunfiletypesizewithflush"]
                    and "rootmerging" not in plugin.lower()
                ):
                    stepToSplit = self.stepsListDict[index]
                    numberOfProdsToInsert = len(stepToSplit["fileTypesOut"])
                    self.prodsTypeList.remove("Merge")
                    self.plugins.pop(index)
                    self.outputSEs.pop(index)
                    self.specialOutputSEs.pop(index)
                    self.priorities.pop(index)
                    self.cpus.pop(index)
                    self.bkQueries.pop(index)
                    self.previousProds.pop(index)
                    self.groupSizes.pop(index)
                    self.removeInputsFlags.pop(index)
                    self.outputFileMasks.pop(index)
                    self.outputFileSteps.pop(index)
                    self.inputs.pop(index)
                    self.inputDataPolicies.pop(index)
                    self.stepsList.pop(index)
                    self.events.pop(index)
                    self.targets.pop(index)
                    self.multicore.pop(index)
                    self.processors.pop(index)
                    self.outputModes.pop(index)
                    self.ancestorDepths.pop(index)
                    newSteps = _splitIntoProductionSteps(stepToSplit)
                    newSteps.reverse()
                    self.stepsListDict.remove(stepToSplit)
                    last = self.stepsInProds.pop(index)[0]
                    for x in range(numberOfProdsToInsert):
                        self.prodsTypeList.insert(index, "Merge")
                        self.plugins.insert(index, plugin)
                        self.outputSEs.insert(index, outputSE)
                        self.specialOutputSEs.insert(index, specialOutputSE)
                        self.priorities.insert(index, priority)
                        self.cpus.insert(index, cpu)
                        self.bkQueries.insert(index, bkQuery)
                        self.groupSizes.insert(index, groupSize)
                        self.removeInputsFlags.insert(index, removeInputsFlag)
                        self.outputFileMasks.insert(index, outputFileMask)
                        self.outputFileSteps.insert(index, outputFileStep)
                        self.inputs.insert(index, inputs)
                        self.inputDataPolicies.insert(index, idp)
                        self.stepsList.insert(index, stepID)
                        self.previousProds.insert(index, preProd)
                        self.stepsListDict.insert(index, newSteps[x])
                        self.stepsInProds.insert(index + x, [last + x])
                        self.events.insert(index, events)
                        self.targets.insert(index, targets)
                        self.multicore.insert(index, multicore)
                        self.processors.insert(index, processors)
                        self.outputModes.insert(index, outputMode)
                        self.ancestorDepths.insert(index, ancestorDepth)

        correctedStepsInProds = []
        toInsert = self.stepsInProds[0][0]
        lengths = [len(x) for x in self.stepsInProds]
        for length in lengths:
            li = [toInsert + x for x in range(length)]
            toInsert += length
            correctedStepsInProds.append(li)

        self.stepsInProds = correctedStepsInProds

    #############################################################################

    def _getProdsDescriptionDict(self):
        """Returns a dictionary representing the description of the request (of all
        the productions in it)"""

        prodsDict = {}

        prodNumber = 1

        for (
            prodType,
            stepsInProd,
            bkQuery,
            removeInputsFlag,
            outputSE,
            priority,
            cpu,
            inputD,
            outputMode,
            outFileMask,
            outFileStep,
            target,
            groupSize,
            plugin,
            idp,
            previousProd,
            events,
            multicore,
            processors,
            ancestorDepth,
        ) in zip(
            self.prodsTypeList,
            self.stepsInProds,
            self.bkQueries,
            self.removeInputsFlags,
            self.outputSEsPerFileType,
            self.priorities,
            self.cpus,
            self.inputs,
            self.outputModes,
            self.outputFileMasks,
            self.outputFileSteps,
            self.targets,
            self.groupSizes,
            self.plugins,
            self.inputDataPolicies,
            self.previousProds,
            self.events,
            self.multicore,
            self.processors,
            self.ancestorDepths,
        ):
            if not self.parentRequestID and self.requestID:
                transformationFamily = self.requestID
            else:
                transformationFamily = self.parentRequestID

            stepsInProdProdNameList = [
                str(self.stepsList[index - 1]) + str(self.stepsListDict[index - 1]["fileTypesIn"])
                for index in stepsInProd
            ]
            prodsDict[prodNumber] = {
                "productionType": prodType,
                "stepsInProd": [self.stepsList[index - 1] for index in stepsInProd],
                "bkQuery": bkQuery,
                "removeInputsFlag": removeInputsFlag,
                "tracking": 0,
                "outputSE": outputSE,
                "priority": priority,
                "cpu": cpu,
                "input": inputD,
                "outputMode": outputMode,
                "outputFileMask": outFileMask,
                "outputFileStep": outFileStep,
                "target": target,
                "groupSize": groupSize,
                "plugin": plugin,
                "inputDataPolicy": idp,
                "derivedProduction": 0,
                "transformationFamily": transformationFamily,
                "previousProd": previousProd,
                "stepsInProd-ProdName": stepsInProdProdNameList,
                "events": events,
                "multicore": multicore,
                "processors": processors,
                "ancestorDepth": ancestorDepth,
            }
            prodNumber += 1

        # tracking the last production(s)
        prodsDict[prodNumber - 1]["tracking"] = 1
        typeOfLastProd = prodsDict[prodNumber - 1]["productionType"]
        index = 2
        try:
            while prodsDict[prodNumber - index]["productionType"] == typeOfLastProd:
                prodsDict[prodNumber - index]["tracking"] = 1
                index += 1
        except KeyError:
            pass

        # production derivation, if necessary
        if self.derivedProduction:
            prodsDict[1]["derivedProduction"] = self.derivedProduction

        return prodsDict

    #############################################################################

    def _buildProduction(
        self,
        prodType,
        stepsInProd,
        outputSE,
        priority,
        cpu,
        inputDataList=None,
        outputMode="Any",
        inputDataPolicy="download",
        outputFileMask="",
        outputFileStep="",
        target="",
        removeInputData=False,
        groupSize=1,
        bkQuery=None,
        plugin="",
        previousProdID=0,
        derivedProdID=0,
        transformationFamily=0,
        events=-1,
        multicore="True",
        processors=(0, 0),
        ancestorDepth=0,
    ):
        """Wrapper around Production API to build a production, given the needed
        parameters.

        Args:
          prodType (str): production type (e.g. 'DataStripping')
          stepsInProd (list): list of steps in the production
          outputSE (dict): dictionary that holds relation between file type and output SE
          priority (int): production priority
          cpu (int): CPU time, in HS06s for jobs in this production

        Returns:
          prod: a Production object
        """
        prod = Production()

        # non optional parameters
        prod.LHCbJob.setType(prodType)
        try:
            fTypeIn = [ft.upper() for ft in stepsInProd[0]["fileTypesIn"]]
        except IndexError:
            fTypeIn = []
        prod.LHCbJob.workflow.setName(
            "Request_%d_%s_%s_EventType_%s_%s_%s"
            % (
                self.requestID,
                prodType,
                self.prodGroup,
                self.eventType,
                "".join([x.split(".")[0] for x in fTypeIn]),
                self.appendName,
            )
        )
        prod.setBKParameters(
            configName=self.outConfigName,
            configVersion=self.configVersion,
            groupDescriptionOrStepsList=stepsInProd,
            conditions=self.dataTakingConditions,
        )
        prod.setParameter("eventType", "string", self.eventType, "Event Type of the production")
        prod.setParameter("numberOfEvents", "string", str(events), "Number of events requested")

        prod.setParameter("multicore", "string", multicore, "Flag for enabling gaudi parallel")
        if processors[0] or processors[1]:
            minProcs = processors[0] if processors[0] else None
            maxProcs = processors[1] if processors[1] else None
            prod.LHCbJob.setNumberOfProcessors(minNumberOfProcessors=minProcs, maxNumberOfProcessors=maxProcs)
        prod.prodGroup = self.prodGroup
        prod.priority = priority
        prod.LHCbJob.workflow.setDescrShort("prodDescription")
        prod.LHCbJob.workflow.setDescription("prodDescription")
        prod.LHCbJob.setCPUTime(cpu)
        prod.plugin = plugin

        # optional parameters
        prod.jobFileGroupSize = groupSize
        if inputDataPolicy:
            prod.LHCbJob.setInputDataPolicy(inputDataPolicy)
        prod.setOutputMode(outputMode)
        if outputFileMask:
            outputFileMask = [m.lower() for m in outputFileMask.replace(" ", "").split(",")]
        if outputFileStep:
            if isinstance(outputFileStep, str):
                outputFileStep = [m.lower() for m in outputFileStep.replace(" ", "").split(",")]
        prod.setFileMask(outputFileMask, outputFileStep)
        if target:
            if target == "Tier2":
                prod.banTier1s()
            elif "BAN" in target:
                sitesToBan = target.split(":")
                if len(sitesToBan) > 1:
                    prod.banSites(sitesToBan[1:])
            elif target != "ALL":
                prod.LHCbJob.setDestination(target)
        prod.LHCbJob.setInputData(inputDataList)
        if ancestorDepth:
            prod.LHCbJob.setAncestorDepth(ancestorDepth)
        if derivedProdID:
            prod.ancestorProduction = derivedProdID
        if transformationFamily:
            prod.transformationFamily = transformationFamily
        if self.fractionToProcess:
            prod.setParameter("FractionToProcess", "string", str(self.fractionToProcess), "Fraction to process")
        if self.minFilesToProcess:
            prod.setParameter("MinFilesToProcess", "string", str(self.minFilesToProcess), "Min N of Files to process")
        if self.processingPass:
            prod.setParameter(
                "processingPass", "string", self.processingPass, "Processing pass of input for the request"
            )

        # Adding optional input BK query
        if bkQuery:
            if bkQuery.lower() == "full":
                prod.inputBKSelection = self._getBKKQuery()
            elif bkQuery.lower() == "frompreviousprod":
                prod.inputBKSelection = self._getBKKQuery("frompreviousprod", fTypeIn, previousProdID)

        self.logger.verbose(f"Launching with BK selection {prod.inputBKSelection}")

        prod = self._addStepsToProd(prod, stepsInProd, removeInputData=removeInputData)

        for ft, oSE in outputSE.items():
            prod.outputSEs.setdefault(ft, oSE)

        prod.LHCbJob.setDIRACPlatform()

        return prod

    #############################################################################

    def _addStepsToProd(self, prod, stepsInProd, stepsSequence="sequential", removeInputData=False):
        """Given a Production object, add requested steps (application and
        finalization)

        Args:
          prod (Production): the Production object to which the steps are added
          stepsInProd (DAG): DAG of steps in a production
          stepsSequence (str or dict): applications steps sequence
          removeInputData (bool): flag that indicates if the input data should be removed (for the finalization step)

        Returns:
          prod with steps added
        """
        # Adding the application steps
        firstStep = stepsInProd.pop(0)
        stepName = prod.addApplicationStep(
            stepDict=firstStep, modulesList=self.modulesList, extraModulesList=self.extraModulesList
        )
        prod.gaudiSteps.append(stepName)

        for step in stepsInProd:
            stepName = prod.addApplicationStep(
                stepDict=step,
                inputData="previousStep",
                modulesList=self.modulesList,
                extraModulesList=self.extraModulesList,
            )
            prod.gaudiSteps.append(stepName)

        # Adding the finalization step
        if removeInputData:
            prod.addFinalizationStep(
                ["UploadOutputData", "RemoveInputData", "UploadLogFile", "UploadMC", "FailoverRequest"]
            )
        else:
            prod.addFinalizationStep()

        return prod

    def _getBKKQuery(self, mode="full", fileType=None, previousProdID=0):
        """simply creates the bkk query dictionary."""

        if fileType is None:
            fileType = []

        if mode.lower() == "full":
            bkQuery = {
                "FileType": ";;;".join(self.bkFileType),
                "EventType": str(self.eventType),
                "ConfigName": self.configName,
                "ConfigVersion": self.configVersion,
            }

            if self.bkSampleMax and self.bkSampleSeed:
                bkQuery["SampleMax"] = self.bkSampleMax
                bkQuery["SampleSeedMD5"] = self.bkSampleSeed

            if self.dataTakingConditions:
                bkQuery["DataTakingConditions"] = self.dataTakingConditions

            if self.processingPass:
                bkQuery["ProcessingPass"] = self.processingPass

            if self.dqFlag:
                bkQuery["DataQualityFlag"] = self.dqFlag.replace(",", ";;;").replace(" ", "")

            if self.extendedDQOK:
                bkQuery["ExtendedDQOK"] = self.extendedDQOK.replace(",", ";;;").replace(" ", "")

            if self.smog2State:
                bkQuery["SMOG2"] = self.smog2State

            if self.startRun and self.runsList or self.endRun and self.runsList:
                raise ValueError("Please don't mix runs list with start/end run")

            if self.endRun and self.startRun:
                if self.endRun < self.startRun:
                    gLogger.error(f"Your end run {self.endRun} should be more than your start run {self.startRun}!")
                    raise ValueError("Error setting start or end run")

            if self.startRun:
                bkQuery["StartRun"] = self.startRun
            if self.endRun:
                bkQuery["EndRun"] = self.endRun

            if self.runsList:
                bkQuery["RunNumbers"] = self.runsList.replace(",", ";;;").replace(" ", "")

            if self.visibility:
                bkQuery["Visible"] = self.visibility

        elif mode.lower() == "frompreviousprod":
            bkQuery = {"FileType": ";;;".join(fileType).replace(" ", ""), "ProductionID": int(previousProdID)}

            if self.eventType:
                bkQuery["EventType"] = str(self.eventType)

            if self.dqFlag:
                bkQuery["DataQualityFlag"] = self.dqFlag.replace(",", ";;;").replace(" ", "")

        return bkQuery

    ################################################################################
    # properties

    def set_stepsList(self, value):
        listInt = []
        i = 0
        while True:
            try:
                listInt.append(int(value[i]))
                i = i + 1
            except ValueError:
                break
            except IndexError:
                break
        self._stepsList = listInt

    def get_stepsList(self):
        return self._stepsList

    stepsList = property(get_stepsList, set_stepsList)

    def set_startRun(self, value):
        if isinstance(value, str):
            value = int(value)
        if value < 0:
            raise ValueError("startRun can not be negative")
        self._startRun = value

    def get_startRun(self):
        return self._startRun

    startRun = property(get_startRun, set_startRun)

    def set_endRun(self, value):
        if isinstance(value, str):
            value = int(value)
        if value < 0:
            raise ValueError("endRun can not be negative")
        self._endRun = value

    def get_endRun(self):
        return self._endRun

    endRun = property(get_endRun, set_endRun)

    def set_requestID(self, value):
        if value == "":
            value = 0
        if isinstance(value, str):
            value = int(value)
        if value < 0:
            raise ValueError("requestID can not be negative")
        self._requestID = value

    def get_requestID(self):
        return self._requestID

    requestID = property(get_requestID, set_requestID)

    def set_parentRequestID(self, value):
        if value == "":
            value = 0
        if isinstance(value, str):
            value = int(value)
        if value < 0:
            raise ValueError("parentRequestID can not be negative")
        self._parentRequestID = value

    def get_parentRequestID(self):
        return self._parentRequestID

    parentRequestID = property(get_parentRequestID, set_parentRequestID)

    def set_bkFileType(self, value):
        if isinstance(value, str):
            value = value.replace(" ", "").split(",")
        self._bkFileType = value

    def get_bkFileType(self):
        return self._bkFileType

    bkFileType = property(get_bkFileType, set_bkFileType)


#############################################################################


def _splitIntoProductionSteps(step):
    """Given a list of bookkeeping steps, produce production steps."""
    prodSteps = []

    if len(step["fileTypesIn"]) <= 1:
        prodSteps.append(step)
    else:
        if set(step["fileTypesOut"]) > set(step["fileTypesIn"]):
            raise ValueError(
                "Step outputs %s are not part of the inputs %s...?"
                % (str(step["fileTypesOut"]), str(step["fileTypesIn"]))
            )
        for outputTypes in step["fileTypesOut"]:
            prodStep = copy.deepcopy(step)
            prodStep["fileTypesIn"] = [outputTypes]
            prodStep["fileTypesOut"] = [outputTypes]
            prodStep["prodStepID"] = str(prodStep["StepId"]) + str(prodStep["fileTypesIn"])
            prodSteps.append(prodStep)

    return prodSteps



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
:mod: StorageUsageHandler

.. module: StorageUsageHandler

:synopsis: Implementation of the Storage Usage service in the DISET framework.
"""
# from DIRAC
from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.DISET.RequestHandler import RequestHandler

# from LHCbDIRAC
from LHCbDIRAC.DataManagementSystem.DB.StorageUsageDB import StorageUsageDB

gStorageUsageDB = False


def initializeStorageUsageHandler(_serviceInfo):
    """handlre initialisation."""
    global gStorageUsageDB
    gStorageUsageDB = StorageUsageDB()
    return S_OK()


class StorageUsageHandler(RequestHandler):
    """
    .. class:: StorageUsageHandler
    """

    types_publishDirectories = [dict]

    @staticmethod
    def export_publishDirectories(directoryDict):
        """export of publishDirectories."""
        return gStorageUsageDB.publishDirectories(directoryDict)

    types_removeDirectory = [str, list, tuple]

    @staticmethod
    def export_removeDirectory(dirPaths):
        """export of removeDirectory."""
        if isinstance(dirPaths, str):
            dirPaths = (dirPaths,)
        for dirPath in dirPaths:
            result = gStorageUsageDB.removeDirectory(dirPath)
            if not result["OK"]:
                gLogger.error("Could not delete directory", f"{dirPath} : {result['Message']}")
                return result
        return S_OK()

    types_removeDirFromSe_Usage = []

    @staticmethod
    def export_removeDirFromSe_Usage(dirPaths):
        """Exports the method to remove entries from the se_Usage table."""
        return gStorageUsageDB.removeDirFromSe_Usage(dirPaths)

    types_removeDirFromProblematicDirs = []

    @staticmethod
    def export_removeDirFromProblematicDirs(dirPaths):
        """Exports the method to remove entries from the problematicDirs table."""
        return gStorageUsageDB.removeDirFromProblematicDirs(dirPaths)

    ##################################################################
    #
    # These are the methods for monitoring the usage
    #

    types_getStorageSummary = []

    @staticmethod
    def export_getStorageSummary(directory="", filetype="", production="", sites=None):
        """Retieve a summary for the storage usage."""
        sites = sites if sites else []
        return gStorageUsageDB.getStorageSummary(directory, filetype, production, sites)

    types_getStorageDirectorySummary = []

    @staticmethod
    def export_getStorageDirectorySummary(directory="", filetype="", production="", sites=None):
        """Retieve a directory summary for the storage usage."""
        sites = sites if sites else []
        result = gStorageUsageDB.getStorageDirectorySummary(directory, filetype, production, sites)
        if not result["OK"]:
            return result
        dl = []
        for dirPath in result["Value"]:
            dl.append((dirPath, result["Value"][dirPath]["Size"], result["Value"][dirPath]["Files"]))
        return S_OK(dl)

    types_getStorageDirectoryData = []

    @staticmethod
    def export_getStorageDirectoryData(directory="", filetype="", production="", sites=None):
        """Retrieve a directory summary for the storage usage."""
        sites = sites if sites else []
        return gStorageUsageDB.getStorageDirectorySummary(directory, filetype, production, sites)

    types_getStorageDirectories = []

    @staticmethod
    def export_getStorageDirectories(directory="", filetype="", production="", sites=None):
        """Retrieve the directories for the supplied selection."""
        sites = sites if sites else []
        return gStorageUsageDB.getStorageDirectories(directory, filetype, production, sites)

    types_getStorageDirectorySummaryWeb = []

    @staticmethod
    def export_getStorageDirectorySummaryWeb(selectDict, sortList, startItem, maxItems):
        """Get the summary of the directory storage summary."""
        resultDict = {}
        # Sorting instructions. Only one for the moment.
        directory = ""
        if "Directory" in selectDict:
            directory = selectDict["Directory"]
        filetype = ""
        if "FileType" in selectDict:
            filetype = selectDict["FileType"]
        production = ""
        if "Production" in selectDict:
            production = selectDict["Production"]
        ses = []
        if "SEs" in selectDict:
            ses = selectDict["SEs"]

        res = gStorageUsageDB.getStorageDirectorySummary(directory, filetype, production, ses)
        if not res["OK"]:
            gLogger.error(
                "StorageUsageHandler.getStorageDirectorySummaryWeb: Failed to obtain directory summary.", res["Message"]
            )
            return res
        dirList = res["Value"]
        dirList = [(path, dirList[path]["Size"], dirList[path]["Files"]) for path in dirList]

        nDirs = len(dirList)
        resultDict["TotalRecords"] = nDirs
        if nDirs == 0:
            return S_OK(resultDict)
        iniDir = startItem
        lastDir = iniDir + maxItems
        if iniDir >= nDirs:
            return S_ERROR("Item number out of range")
        if lastDir > nDirs:
            lastDir = nDirs

        # prepare the extras count
        res = gStorageUsageDB.getStorageSummary(directory, filetype, production, ses)
        if not res["OK"]:
            gLogger.error(
                "StorageUsageHandler.getStorageDirectorySummaryWeb: Failed to obtain usage summary.", res["Message"]
            )
            return res
        resultDict["Extras"] = res["Value"]

        # prepare the standard structure now
        resultDict["ParameterNames"] = ["Directory Path", "Size", "Files"]
        resultDict["Records"] = dirList[iniDir:lastDir]
        return S_OK(resultDict)

    types_getStorageElementSelection = []

    @staticmethod
    def export_getStorageElementSelection():
        """Retrieve the possible selections."""
        return gStorageUsageDB.getStorageElementSelection()

    types_getUserStorageUsage = []

    @staticmethod
    def export_getUserStorageUsage(userName=False):
        """Retrieve a summary of the user usage."""
        return gStorageUsageDB.getUserStorageUsage(userName)

    types_getUserSummaryPerSE = []

    @staticmethod
    def export_getUserSummaryPerSE(userName=False):
        """Retrieve a summary of the user usage per SE."""
        return gStorageUsageDB.getUserSummaryPerSE(userName)

    types_getDirectorySummaryPerSE = []

    @staticmethod
    def export_getDirectorySummaryPerSE(directory):
        """Retrieve a summary (total files and total size) for a given directory,
        grouped by storage element."""
        return gStorageUsageDB.getDirectorySummaryPerSE(directory)

    types_getRunSummaryPerSE = []

    @staticmethod
    def export_getRunSummaryPerSE(run):
        """Retrieve a summary (total files and total size) for a given run, grouped
        by storage element."""
        return gStorageUsageDB.getRunSummaryPerSE(run)

    types_getIDs = []

    @staticmethod
    def export_getIDs(dirList):
        """Check if the directories exist in the su_Directory table and if yes
        returns the IDs."""
        return gStorageUsageDB.getIDs(dirList)

    types_getAllReplicasInFC = []

    @staticmethod
    def export_getAllReplicasInFC(path):
        """Export the DB method to query the su_seUsage table to get all the
        entries relative to a given path registered in the FC.

        Returns for every replica the SE, the update, the files and the size
        """
        return gStorageUsageDB.getAllReplicasInFC(path)

    ####
    # Catalog
    ####

    types_getSummary = [str]

    @staticmethod
    def export_getSummary(path, fileType=False, production=False):
        """export of getSummary."""
        return gStorageUsageDB.getSummary(path, fileType, production)

    types_getUserSummary = []

    @staticmethod
    def export_getUserSummary(userName=False):
        """export of getUserSummary."""
        return gStorageUsageDB.getUserSummary(userName)

    ####
    # Purge
    ####

    types_purgeOutdatedEntries = [str, int]

    @staticmethod
    def export_purgeOutdatedEntries(rootDir, outdatedSeconds, preserveDirsList=None):
        """Purge entries that haven't been updated in the last outdated seconds."""
        preserveDirsList = preserveDirsList if preserveDirsList else []
        return gStorageUsageDB.purgeOutdatedEntries(rootDir, outdatedSeconds, preserveDirsList)

    ###
    # methods to deal with problematicDirs directory: problematicDirs
    ###
    types_publishToProblematicDirs = []

    @staticmethod
    def export_publishToProblematicDirs(directoryDict):
        """Export the publishToProblematicDirs DB method, which inserts/updates row
        into the  problematicDirs."""
        return gStorageUsageDB.publishToProblematicDirs(directoryDict)

    types_getProblematicDirsSummary = []

    @staticmethod
    def export_getProblematicDirsSummary(site, problem=False):
        """Exports the getProblematicDirsSummary method: returns a list of
        directories from the problematicDirs table, that have some inconsistency
        between the SE dumps and the FC."""
        return gStorageUsageDB.getProblematicDirsSummary(site, problem)

    types_removeAllFromProblematicDirs = []

    @staticmethod
    def export_removeAllFromProblematicDirs(site=False):
        """Exports the removeAllFromProblematicDirs method: delete all entries from
        problematicDirs table for a give site (optional argument)"""
        return gStorageUsageDB.removeAllFromProblematicDirs(site)

    ###
    # methods to deal with se_Usage table
    ###
    types_publishToSEReplicas = []

    @staticmethod
    def export_publishToSEReplicas(directoryDict):
        """Export the publishToSEReplicas DB method, which inserts/updates replicas
        on the SE to the se_Usage table."""
        return gStorageUsageDB.publishToSEReplicas(directoryDict)

    ###
    # methods to deal with se_STSummary table
    ###

    types_publishTose_STSummary = []

    @staticmethod
    def export_publishTose_STSummary(site, spaceToken, totalSize, totalFiles, StorageDumpUpdate):
        """Export the publishTose_STSummary DB method, which inserts/updates the
        reports of total files and total size from the storage dumps to the
        se_STSummary table."""
        return gStorageUsageDB.publishTose_STSummary(site, spaceToken, totalSize, totalFiles, StorageDumpUpdate)

    types_getSTSummary = []

    @staticmethod
    def export_getSTSummary(site, spaceToken=False):
        """Exports getSTSummary method: returns a summary of the used space for the
        given site, based on the storage dumps provided by sites."""
        return gStorageUsageDB.getSTSummary(site, spaceToken)

    types_removeSTSummary = []

    @staticmethod
    def export_removeSTSummary(site, spaceToken=False):
        """Exports removeSTSummary method: removes from the se_STSummary table all
        entries relative to the given site and (optionally ) space token."""
        return gStorageUsageDB.removeSTSummary(site, spaceToken)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
""" Module containing a front-end to the ElasticSearch-based ElasticPrMonDB.
"""

from DIRAC import gConfig
from DIRAC.ConfigurationSystem.Client.PathFinder import getDatabaseSection
from DIRAC.ConfigurationSystem.Client.Helpers import CSGlobals
from LHCbDIRAC.ProductionManagementSystem.DB.ElasticMCStatsDBBase import ElasticMCStatsDBBase

name = "ElasticPrMonDB"

mapping = {
    "properties": {
        "wmsID": {"type": "long"},
        "ProductionID": {"type": "integer"},
        "JobID": {"type": "integer"},
        "Avg": {
            "properties": {
                "nprocs": {"type": "float"},
                "nthreads": {"type": "float"},
                "pss": {"type": "float"},
                "rchar": {"type": "float"},
                "read_bytes": {"type": "float"},
                "rss": {"type": "float"},
                "rx_bytes": {"type": "float"},
                "rx_packets": {"type": "float"},
                "swap": {"type": "float"},
                "tx_bytes": {"type": "float"},
                "tx_packets": {"type": "float"},
                "vmem": {"type": "float"},
                "wchar": {"type": "float"},
                "write_bytes": {"type": "float"},
            }
        },
        "Max": {
            "properties": {
                "nprocs": {"type": "integer"},
                "nthreads": {"type": "integer"},
                "pss": {"type": "long"},
                "rchar": {"type": "long"},
                "read_bytes": {"type": "long"},
                "rss": {"type": "long"},
                "rx_bytes": {"type": "long"},
                "rx_packets": {"type": "long"},
                "swap": {"type": "integer"},
                "tx_bytes": {"type": "long"},
                "tx_packets": {"type": "long"},
                "vmem": {"type": "long"},
                "wchar": {"type": "long"},
                "write_bytes": {"type": "long"},
                "stime": {"type": "integer"},
                "utime": {"type": "integer"},
                "wtime": {"type": "integer"},
            }
        },
        "HW": {
            "properties": {
                "cpu": {
                    "properties": {
                        "CPUs": {"type": "integer"},
                        "CoresPerSocket": {"type": "integer"},
                        "ModelName": {"type": "text"},
                        "Sockets": {"type": "integer"},
                        "ThreadsPerCore": {"type": "integer"},
                    }
                },
                "mem": {"properties": {"MemTotal": {"type": "long"}}},
            }
        },
    }
}


class ElasticPrMonDB(ElasticMCStatsDBBase):
    def __init__(self):
        """Standard Constructor"""

        section = getDatabaseSection("ProductionManagement", "ElasticPrMonDB")
        indexPrefix = gConfig.getValue(f"{section}/IndexPrefix", CSGlobals.getSetup()).lower()

        # Connecting to the ES cluster
        super().__init__(name, "ProductionManagement/ElasticPrMonDB", indexPrefix)

        self.indexName = f"{self.getIndexPrefix()}_{name.lower()}"
        # Verifying if the index is there, and if not create it
        if not self.client.indices.exists(self.indexName):
            result = self.createIndex(self.indexName, mapping, period=None)
            if not result["OK"]:
                self.log.error(result["Message"])
                raise RuntimeError(result["Message"])
            self.log.always("Index created:", self.indexName)

        self.dslSearch = self._Search(self.indexName)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""A module defining the state machine for the Productions."""
from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.ResourceStatusSystem.PolicySystem.StateMachine import State, StateMachine


class LHCbStateMachine(StateMachine):
    """Just redefining one method."""

    def setState(self, candidateState):
        """Makes sure the state is either None or known to the machine, and that it
        is a valid state to move into.

        Final states are also checked. This is a re-definition of original
        one that wasn't making these checks
        """

        if candidateState == self.state:
            return S_OK(candidateState)

        if candidateState is None:
            self.state = candidateState
        elif candidateState in self.states:
            if not self.states[self.state].stateMap:
                gLogger.warn("Final state, won't move")
                return S_OK(self.state)
            if candidateState not in self.states[self.state].stateMap:
                gLogger.warn(f"Can't move from {self.state} to {candidateState}, choosing a good one")
            nextState = self.getNextState(candidateState)
            if not nextState["OK"]:
                return nextState
            nextState = nextState["Value"]
            # If the StateMachine does not accept the candidate, return error message
            self.state = nextState
        else:
            return S_ERROR(f"{candidateState} is not a valid state")

        return S_OK(nextState)


class ProductionsStateMachine(LHCbStateMachine):
    """PMS (Production Management System, not what google thinks!) implementation
    of the state machine."""

    def __init__(self, state):
        """c'tor Defines the state machine transactions."""

        super().__init__(state)

        # Current states
        self.states = {
            "Cleaned": State(15),  # final state
            "Cleaning": State(14, ["Cleaned", "TransformationCleaned"], defState="Cleaning"),
            "Completing": State(13, ["Validating"], defState="Completing"),
            "Stopped": State(12, ["Active", "Completed"], defState="Stopped"),
            "TransformationCleaned": State(16, ["Archived", "Cleaned"], defState="Archived"),
            "Archived": State(11),  # final state
            "Completed": State(10, ["TransformationCleaned", "Archived"], defState="Completed"),
            "WaitingIntegrity": State(9, ["ValidatedOutput"], defState="WaitingIntegrity"),
            "ValidatedOutput": State(8, ["Active", "Completed"], defState="ValidatedOutput"),
            "ValidatingOutput": State(7, ["Active", "ValidatedOutput", "WaitingIntegrity"], defState="ValidatedOutput"),
            "RemovedFiles": State(6, ["Completed"], defState="RemovedFiles"),
            "RemovingFiles": State(5, ["RemovedFiles"], defState="RemovingFiles"),
            "ValidatingInput": State(4, ["Active", "RemovingFiles"], defState="ValidatingInput"),
            "Flush": State(3, ["Active"], defState="Active"),
            "Idle": State(
                2,
                ["Active", "Stopped", "ValidatingInput", "ValidatingOutput", "Completed", "Cleaning", "Testing"],
                defState="Idle",
            ),
            "Active": State(
                1,
                [
                    "Flush",
                    "Idle",
                    "Stopped",
                    "Completing",
                    "ValidatingInput",
                    "ValidatingOutput",
                    "TransformationCleaned",
                ],
                defState="Flush",
            ),
            "Testing": State(17, ["Idle"], defState="Testing"),
            "New": State(
                0, ["Active", "Testing", "Cleaning", "TransformationCleaned"], defState="New"
            ),  # initial state
        }

        # NEW states proposal
        # self.states = {
        #     'Cleaned': State(10), # final state
        #     'Cleaning': State(9, ['Cleaned']),
        #     'Completing': State(8, ['Validating', 'Cleaning'], defState = 'Validating'),
        #     'TransformationCleaned': State(16, ['Archived', 'Cleaned'], defState = 'Archived'),
        #     'Stopped': State(7, ['Active', 'Flush', 'Cleaning'], defState = 'Active'),
        #     'Archived': State(6),  # final state
        #     'Completed': State(5, ['Archived'], defState = 'Archived'),
        #     'Validating': State(4, ['Active', 'Completed', 'Cleaning'], defState = 'Completed'),
        #     # Idle to Completed? Should it be that when setting a prod to completed it actually goes to validating?
        #     'Idle': State(3, ['Active', 'Validating', 'Cleaning'], defState = 'Active'),
        #     'Flush': State(2, ['Active', 'Cleaning'], defState = 'Active'),
        #     'Active': State(1, ['Flush', 'Idle', 'Stopped', 'Completing', 'Validating', 'Cleaning'], defState = 'Flush'),
        #     'New': State(0, ['Active', 'Cleaning'], defState = 'Active')  # initial state
        # }



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Debug files status for a (list of) transformations It is possible to do
minor fixes to those files, using options."""

from DIRAC.Core.Base.Script import Script


@Script()
def main():
    from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript

    infoList = ("files", "runs", "tasks", "jobs", "alltasks", "flush", "log", "timing")
    statusList = (
        "Unused",
        "Assigned",
        "Done",
        "Problematic",
        "MissingInFC",
        "MaxReset",
        "Processed",
        "NotProcessed",
        "Removed",
        "ProbInFC",
    )
    dmScript = DMScript()
    dmScript.registerFileSwitches()
    Script.registerSwitch("", "Info=", f"Specify what to print out from {str(infoList)}")
    Script.registerSwitch("", "Status=", f"Select files with a given status from {str(statusList)}")
    Script.registerSwitch("", "Since=", "Select files since a given date (yyyy-mm-dd)")
    Script.registerSwitch("", "Runs=", "Specify a (list of) runs")
    Script.registerSwitch(
        "", "ExceptActiveRunsFromProduction=", "Do not consider runs with active files from these prods"
    )
    Script.registerSwitch("", "SEs=", "Specify a (list of) target SEs")
    Script.registerSwitch("", "Tasks=", "Specify a (list of) tasks")
    Script.registerSwitch("", "Jobs=", "Specify a (list of) jobs")
    Script.registerSwitch("", "DumpFiles", "Dump the list of LFNs on stdout")
    Script.registerSwitch("", "Statistics", "Get the statistics of tasks per status and SE")
    Script.registerSwitch("", "FixRun", "Fix the run number in transformation table")
    Script.registerSwitch("", "FixIt", "Fix the FC")
    Script.registerSwitch("", "KickRequests", "Reset old Assigned requests to Waiting")
    Script.registerSwitch("", "CheckWaitingTasks", "Check if waiting tasks are failed, done or orphan")
    Script.registerSwitch("", "CheckSubmittedTasks", "Check Submitted tasks that do not have an external ID")
    Script.registerSwitch("v", "Verbose", "")
    Script.setUsageMessage(
        "\n".join([__doc__, "Usage:", "  dirac-transformation-debug [options] transID[,transID2[,transID3[,...]]]"])
    )

    Script.parseCommandLine(ignoreErrors=True)

    from LHCbDIRAC.TransformationSystem.Client.TransformationDebug import TransformationDebug

    TransformationDebug().debugTransformation(dmScript, infoList, statusList)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Get the list of all the user files."""

from DIRAC.Core.Base.Script import Script


@Script()
def main():
    from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript

    days = 0
    months = 0
    years = 0
    depth = 1
    wildcard = "*"
    dmScript = DMScript()
    dmScript.registerNamespaceSwitches()
    Script.registerSwitch("", "Days=", f"Match files older than number of days [{days}]")
    Script.registerSwitch("", "Months=", f"Match files older than number of months [{months}]")
    Script.registerSwitch("", "Years=", f"Match files older than number of years [{years}]")
    Script.registerSwitch("", "Wildcard=", f"Wildcard for matching filenames [{wildcard}]")
    Script.registerSwitch("", "Output", "Write list to an output file")
    Script.registerSwitch("", "EmptyDirs", "Create a list of empty directories")
    Script.registerSwitch("", "Depth=", "Depth to which recursively browse (default = %d)" % depth)
    Script.registerSwitch("r", "Recursive", "Set depth to infinite")
    Script.registerSwitch("", "NoDirectories", "Only print out only files, not subdirectories")
    dmScript.registerBKSwitches()

    Script.setUsageMessage(
        __doc__
        + "\n".join(
            [
                "Usage:",
                f"  {Script.scriptName} [option|cfgfile] ...",
            ]
        )
    )

    Script.parseCommandLine(ignoreErrors=False)
    from LHCbDIRAC.DataManagementSystem.Client.ScriptExecutors import executeListDirectory
    from DIRAC import exit

    exit(executeListDirectory(dmScript, days, months, years, wildcard, depth))


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Unittest for:
LHCbDIRAC.AccountingSystem.private.Plotters.DataStoragePlotter.

DataStoragePlotter.__bases__:
  DIRAC.AccountingSystem.private.Plotters.BaseReporter

We are assuming there is a solid test of __bases__, we are not testing them
here and assuming they work fine.

IMPORTANT: the test MUST be pylint compliant !
"""
# pylint: disable=protected-access

import unittest
import math
import operator

from decimal import Decimal

from unittest import mock

from PIL import Image
from functools import reduce


def compare(file1Path, file2Path):
    """Function used to compare two plots.

    returns 0.0 if both are identical
    """

    # Crops image to remove the "Generated on xxxx UTC" string
    image1 = Image.open(file1Path).crop((0, 0, 800, 570))
    image2 = Image.open(file2Path).crop((0, 0, 800, 570))
    h1 = image1.histogram()
    h2 = image2.histogram()
    rms = math.sqrt(reduce(operator.add, map(lambda a, b: (a - b) ** 2, h1, h2)) / len(h1))
    return rms


# ...............................................................................


class DataStoragePlotterTestCase(unittest.TestCase):
    """DataStoragePlotterTestCase."""

    moduleTested = None
    classsTested = None

    def mockModuleTested(self, moduleTested):
        """Used to not redo the mocking done on the parent class ( if any )"""

        # Tries to get the mocks of the parent TestCases ( if any )
        for baseClass in self.__class__.__bases__:
            try:
                # pylint: disable=no-member
                moduleTested = baseClass.mockModuleTested(moduleTested)
            except TypeError:
                continue

        # And then makes its own mock
        class MockDataStorage:
            # pylint: disable=missing-docstring,no-init
            definitionKeyFields = (
                "DataType",
                "Activity",
                "FileType",
                "Production",
                "ProcessingPass",
                "Conditions",
                "EventType",
                "StorageElement",
            )

        moduleTested.DataStorage = mock.Mock(return_value=MockDataStorage())

        return moduleTested

    def setUp(self):
        """Setup the test case."""

        import LHCbDIRAC.AccountingSystem.private.Plotters.DataStoragePlotter as moduleTested

        self.moduleTested = self.mockModuleTested(moduleTested)
        self.classsTested = self.moduleTested.DataStoragePlotter

    def tearDown(self):
        """Tear down the test case."""

        del self.moduleTested
        del self.classsTested


# ...............................................................................


class DataStoragePlotterUnitTest(DataStoragePlotterTestCase):
    """DataStoragePlotterUnitTest.

    <constructor>
     - test_instantiate
    <class variables>
     - test_typeName
     - test_typeKeyFields
     - test_noSEtypeKeyFields
     - test_noSEGrouping
     - test_reportCatalogSpaceName
     - test_reportCatalogFilesName
     - test_reportPhysicalSpaceName
     - test_reportPhysicalFilesName
    <methods>
     - test_reportCatalogSpace
     - test_reportCatalogFiles
     - test_reportPhysicalSpace
     - test_reportPhysicalFiles
     - test_plotCatalogSpace
     - test_plotCatalogFiles
     - test_plotPhysicalSpace
     - test_plotPhysicalFiles
    """

    def test_instantiate(self):
        """tests that we can instantiate one object of the tested class."""
        obj = self.classsTested(None, None)
        self.assertEqual("DataStoragePlotter", obj.__class__.__name__)

    def test_typeName(self):
        """test the class variable "_typeName"."""
        obj = self.classsTested(None, None)
        self.assertEqual(obj._typeName, "DataStorage")

    #   def test_typeKeyFields( self ):
    #     ''' test the class variable "_typeKeyFields"
    #     '''
    #     obj = self.classsTested( None, None )
    #     self.assertEqual( obj._typeKeyFields, [ 'DataType', 'Activity', 'FileType',
    #                                             'Production', 'ProcessingPass',
    #                                             'Conditions', 'EventType', 'StorageElement' ] )

    def test_reportCatalogSpaceName(self):
        """test the class variable "_reportCatalogSpaceName"."""
        obj = self.classsTested(None, None)
        self.assertEqual(obj._reportCatalogSpaceName, "LFN size")

    def test_reportCatalogFilesName(self):
        """test the class variable "_reportCatalogFilesName"."""
        obj = self.classsTested(None, None)
        self.assertEqual(obj._reportCatalogFilesName, "LFN files")

    def test_reportPhysicalSpaceName(self):
        """test the class variable "_reportPhysicalSpaceName"."""
        obj = self.classsTested(None, None)
        self.assertEqual(obj._reportPhysicalSpaceName, "PFN size")

    def test_reportPhysicalFilesName(self):
        """test the class variable "_reportPhysicalFilesName"."""
        obj = self.classsTested(None, None)
        self.assertEqual(obj._reportPhysicalFilesName, "PFN files")

    def test_reportCatalogSpace(self):
        """test the method "_reportCatalogSpace"."""

        mockAccountingDB = mock.Mock()
        mockAccountingDB._getConnection.return_value = {"OK": False, "Message": "No connection"}
        mockAccountingDB.retrieveBucketedData.return_value = {"OK": True, "Value": []}
        mockAccountingDB.calculateBucketLengthForTime.return_value = "BucketLength"
        obj = self.classsTested(mockAccountingDB, None)

        res = obj._reportCatalogSpace({"grouping": "StorageElement"})
        self.assertEqual(res["OK"], False, msg="Rejected StorageElement grouping")
        self.assertEqual(res["Message"], "Grouping by storage element when requesting lfn info makes no sense")

        res = obj._reportCatalogSpace(
            {
                "grouping": "NextToABeer",
                "groupingFields": [0, ["mehh"], "blah"],
                "startTime": "startTime",
                "endTime": "endTime",
                "condDict": {},
            }
        )
        #     self.assertEqual( res[ 'OK' ], False )
        #     self.assertEqual( res[ 'Message' ], 'No connection' )

        # Changed mocked to run over different lines of code
        mockAccountingDB._getConnection.return_value = {"OK": True, "Value": []}
        res = obj._reportCatalogSpace(
            {
                "grouping": "NextToABeer",
                "groupingFields": [0, ["mehh"], "blah"],
                "startTime": "startTime",
                "endTime": "endTime",
                "condDict": {},
            }
        )
        self.assertEqual(res["OK"], True)
        self.assertEqual(res["Value"], {"graphDataDict": {}, "data": {}, "unit": "MB", "granularity": "BucketLength"})

        mockedData = (
            ("Full stream", 1355616000, 86400, Decimal("935388524246.91630989384787")),
            ("Full stream", 1355702400, 86400, Decimal("843844487074.82197482051816")),
        )
        mockAccountingDB.retrieveBucketedData.return_value = {"OK": True, "Value": mockedData}
        mockAccountingDB.calculateBucketLengthForTime.return_value = 86400

        res = obj._reportCatalogSpace(
            {
                "grouping": "EventType",
                "groupingFields": ("%s", ["EventType"]),
                "startTime": 1355663249.0,
                "endTime": 1355749690.0,
                "condDict": {"EventType": "Full stream"},
            }
        )
        self.assertEqual(res["OK"], True)
        self.assertEqual(
            res["Value"],
            {
                "graphDataDict": {"Full stream": {1355616000: 935.38852424691629, 1355702400: 843.84448707482204}},
                "data": {"Full stream": {1355616000: 935388.52424691629, 1355702400: 843844.48707482207}},
                "unit": "GB",
                "granularity": 86400,
            },
        )

    def test_reportCatalogFiles(self):
        """test the method "_reportCatalogFiles"."""

        mockAccountingDB = mock.Mock()
        mockAccountingDB._getConnection.return_value = {"OK": False, "Message": "No connection"}
        mockAccountingDB.retrieveBucketedData.return_value = {"OK": True, "Value": []}
        mockAccountingDB.calculateBucketLengthForTime.return_value = "BucketLength"
        obj = self.classsTested(mockAccountingDB, None)

        res = obj._reportCatalogFiles({"grouping": "StorageElement"})
        self.assertEqual(res["OK"], False)
        self.assertEqual(res["Message"], "Grouping by storage element when requesting lfn info makes no sense")

        res = obj._reportCatalogFiles(
            {
                "grouping": "NextToABeer",
                "groupingFields": [0, ["mehh"], "blah"],
                "startTime": "startTime",
                "endTime": "endTime",
                "condDict": {},
            }
        )
        #     self.assertEqual( res[ 'OK' ], False )
        #     self.assertEqual( res[ 'Message' ], 'No connection' )

        # Changed mocked to run over different lines of code
        mockAccountingDB._getConnection.return_value = {"OK": True, "Value": []}
        res = obj._reportCatalogFiles(
            {
                "grouping": "NextToABeer",
                "groupingFields": [0, ["mehh"], "blah"],
                "startTime": "startTime",
                "endTime": "endTime",
                "condDict": {},
            }
        )
        self.assertEqual(res["OK"], True, msg="Expected S_OK")
        self.assertEqual(
            res["Value"], {"graphDataDict": {}, "data": {}, "unit": "files", "granularity": "BucketLength"}
        )

        mockedData = (
            ("Full stream", 1355616000, 86400, Decimal("420.47885754501202")),
            ("Full stream", 1355702400, 86400, Decimal("380.35170637810842")),
        )
        mockAccountingDB.retrieveBucketedData.return_value = {"OK": True, "Value": mockedData}
        mockAccountingDB.calculateBucketLengthForTime.return_value = 86400

        res = obj._reportCatalogFiles(
            {
                "grouping": "EventType",
                "groupingFields": ("%s", ["EventType"]),
                "startTime": 1355663249.0,
                "endTime": 1355749690.0,
                "condDict": {"EventType": "Full stream"},
            }
        )
        self.assertEqual(res["OK"], True)
        self.assertEqual(
            res["Value"],
            {
                "graphDataDict": {"Full stream": {1355616000: 420.47885754501203, 1355702400: 380.35170637810842}},
                "data": {"Full stream": {1355616000: 420.47885754501203, 1355702400: 380.35170637810842}},
                "unit": "files",
                "granularity": 86400,
            },
        )

    def test_reportPhysicalSpace(self):
        """test the method "_reportPhysicalSpace"."""

        mockAccountingDB = mock.Mock()
        mockAccountingDB._getConnection.return_value = {"OK": False, "Message": "No connection"}
        mockAccountingDB.retrieveBucketedData.return_value = {"OK": True, "Value": []}
        mockAccountingDB.calculateBucketLengthForTime.return_value = "BucketLength"
        obj = self.classsTested(mockAccountingDB, None)

        res = obj._reportPhysicalSpace(
            {"groupingFields": [0, ["mehh"], "blah"], "startTime": "startTime", "endTime": "endTime", "condDict": {}}
        )
        #     self.assertEqual( res[ 'OK' ], False )
        #     self.assertEqual( res[ 'Message' ], 'No connection' )

        # Changed mocked to run over different lines of code
        mockAccountingDB._getConnection.return_value = {"OK": True, "Value": []}
        res = obj._reportPhysicalSpace(
            {
                "grouping": "NextToABeer",
                "groupingFields": [0, ["mehh"], "blah"],
                "startTime": "startTime",
                "endTime": "endTime",
                "condDict": {},
            }
        )
        self.assertEqual(res["OK"], True, msg="Expected S_OK")
        self.assertEqual(res["Value"], {"graphDataDict": {}, "data": {}, "unit": "MB", "granularity": "BucketLength"})

        mockedData = (
            ("Full stream", 1355616000, 86400, Decimal("14754501.202")),
            ("Full stream", 1355702400, 86400, Decimal("15237810.842")),
        )
        mockAccountingDB.retrieveBucketedData.return_value = {"OK": True, "Value": mockedData}
        mockAccountingDB.calculateBucketLengthForTime.return_value = 86400

        res = obj._reportPhysicalSpace(
            {
                "grouping": "EventType",
                "groupingFields": ("%s", ["EventType"]),
                "startTime": 1355663249.0,
                "endTime": 1355749690.0,
                "condDict": {"EventType": "Full stream"},
            }
        )
        self.assertEqual(res["OK"], True)
        self.assertEqual(
            res["Value"],
            {
                "graphDataDict": {"Full stream": {1355616000: 14.754501202, 1355702400: 15.237810842}},
                "data": {"Full stream": {1355616000: 14.754501202, 1355702400: 15.237810842}},
                "unit": "MB",
                "granularity": 86400,
            },
        )

    def test_reportPhysicalFiles(self):
        """test the method "_reportPhysicalFiles"."""

        mockAccountingDB = mock.Mock()
        mockAccountingDB._getConnection.return_value = {"OK": False, "Message": "No connection"}
        mockAccountingDB.retrieveBucketedData.return_value = {"OK": True, "Value": []}
        mockAccountingDB.calculateBucketLengthForTime.return_value = "BucketLength"
        obj = self.classsTested(mockAccountingDB, None)

        res = obj._reportPhysicalFiles(
            {"groupingFields": [0, ["mehh"], "blah"], "startTime": "startTime", "endTime": "endTime", "condDict": {}}
        )
        #     self.assertEqual( res[ 'OK' ], False )
        #     self.assertEqual( res[ 'Message' ], 'No connection' )

        # Changed mocked to run over different lines of code
        mockAccountingDB._getConnection.return_value = {"OK": True, "Value": []}
        res = obj._reportPhysicalFiles(
            {
                "grouping": "NextToABeer",
                "groupingFields": [0, ["mehh"], "blah"],
                "startTime": "startTime",
                "endTime": "endTime",
                "condDict": {},
            }
        )
        self.assertEqual(res["OK"], True, msg="Expected S_OK")
        self.assertEqual(
            res["Value"], {"graphDataDict": {}, "data": {}, "unit": "files", "granularity": "BucketLength"}
        )

        mockedData = (
            ("Full stream", 1355616000, 86400, Decimal("42.47885754501202")),
            ("Full stream", 1355702400, 86400, Decimal("38.35170637810842")),
        )
        mockAccountingDB.retrieveBucketedData.return_value = {"OK": True, "Value": mockedData}
        mockAccountingDB.calculateBucketLengthForTime.return_value = 86400

        res = obj._reportPhysicalFiles(
            {
                "grouping": "EventType",
                "groupingFields": ("%s", ["EventType"]),
                "startTime": 1355663249.0,
                "endTime": 1355749690.0,
                "condDict": {"EventType": "Full stream"},
            }
        )
        self.assertEqual(res["OK"], True)
        self.assertEqual(
            res["Value"],
            {
                "graphDataDict": {"Full stream": {1355616000: 42.47885754501202, 1355702400: 38.35170637810842}},
                "data": {"Full stream": {1355616000: 42.47885754501202, 1355702400: 38.35170637810842}},
                "unit": "files",
                "granularity": 86400,
            },
        )


# ...............................................................................


class DataStoragePlotterUnitTestCrashes(DataStoragePlotterTestCase):
    """DataStoragePlotterUnitTestCrashes.

    <constructor>
     - test_instantiate
    <class variables>
    <methods>
     - test_reportCatalogSpace
     - test_reportCatalogFiles
     - test_reportPhysicalSpace
     - test_reportPhysicalFiles
     - test_plotCatalogSpace
     - test_plotCatalogFiles
     - test_plotPhysicalSpace
     - test_plotPhysicalFiles
    """

    def test_instantiate(self):
        """test the constructor."""

        self.assertRaises(TypeError, self.classsTested)
        self.assertRaises(TypeError, self.classsTested, None)
        self.assertRaises(TypeError, self.classsTested, None, None, None, None)

        self.assertRaises(TypeError, self.classsTested, extraArgs=None)
        self.assertRaises(TypeError, self.classsTested, None, extraArgs=None)
        self.assertRaises(TypeError, self.classsTested, None, None, None, extraArgs=None)

    def test_reportCatalogSpace(self):
        """test the method "_reportCatalogSpace"."""

        mockAccountingDB = mock.Mock()
        mockAccountingDB._getConnection.return_value = {"OK": False, "Message": "No connection"}
        mockAccountingDB.retrieveBucketedData.return_value = {"OK": True, "Value": []}
        mockAccountingDB.calculateBucketLengthForTime.return_value = "BucketLength"
        obj = self.classsTested(mockAccountingDB, None)

        self.assertRaises(KeyError, obj._reportCatalogSpace, {})
        self.assertRaises(KeyError, obj._reportCatalogSpace, {"grouping": 1})
        self.assertRaises(IndexError, obj._reportCatalogSpace, {"grouping": 1, "groupingFields": []})
        self.assertRaises(TypeError, obj._reportCatalogSpace, {"grouping": 1, "groupingFields": [1, 2]})
        self.assertRaises(TypeError, obj._reportCatalogSpace, {"grouping": 1, "groupingFields": [1, [2]]})
        self.assertRaises(TypeError, obj._reportCatalogSpace, {"grouping": 1, "groupingFields": ["1", "2"]})
        self.assertRaises(KeyError, obj._reportCatalogSpace, {"grouping": 1, "groupingFields": ["1", [2]]})
        self.assertRaises(
            KeyError, obj._reportCatalogSpace, {"grouping": 1, "groupingFields": ["1", [2, 2]], "startTime": None}
        )
        self.assertRaises(
            KeyError,
            obj._reportCatalogSpace,
            {"grouping": 1, "groupingFields": ["1", [2, 2]], "startTime": None, "endTime": None},
        )
        self.assertRaises(
            TypeError,
            obj._reportCatalogSpace,
            {"grouping": 1, "groupingFields": ["1", [2, 2]], "startTime": None, "endTime": None, "condDict": None},
        )

    def test_reportCatalogFiles(self):
        """test the method "_reportCatalogFiles"."""

        mockAccountingDB = mock.Mock()
        mockAccountingDB._getConnection.return_value = {"OK": False, "Message": "No connection"}
        mockAccountingDB.retrieveBucketedData.return_value = {"OK": True, "Value": []}
        mockAccountingDB.calculateBucketLengthForTime.return_value = "BucketLength"
        obj = self.classsTested(mockAccountingDB, None)

        self.assertRaises(KeyError, obj._reportCatalogFiles, {})
        self.assertRaises(KeyError, obj._reportCatalogFiles, {"grouping": 1})
        self.assertRaises(IndexError, obj._reportCatalogFiles, {"grouping": 1, "groupingFields": []})
        self.assertRaises(TypeError, obj._reportCatalogFiles, {"grouping": 1, "groupingFields": [1, 2]})
        self.assertRaises(TypeError, obj._reportCatalogFiles, {"grouping": 1, "groupingFields": [1, [2]]})
        self.assertRaises(TypeError, obj._reportCatalogFiles, {"grouping": 1, "groupingFields": ["1", "2"]})
        self.assertRaises(KeyError, obj._reportCatalogFiles, {"grouping": 1, "groupingFields": ["1", [2]]})
        self.assertRaises(
            KeyError, obj._reportCatalogFiles, {"grouping": 1, "groupingFields": ["1", [2, 2]], "startTime": None}
        )
        self.assertRaises(
            KeyError,
            obj._reportCatalogFiles,
            {"grouping": 1, "groupingFields": ["1", [2, 2]], "startTime": None, "endTime": None},
        )
        self.assertRaises(
            TypeError,
            obj._reportCatalogFiles,
            {"grouping": 1, "groupingFields": ["1", [2, 2]], "startTime": None, "endTime": None, "condDict": None},
        )

    def test_reportPhysicalSpace(self):
        """test the method "_reportPhysicalSpace"."""

        mockAccountingDB = mock.Mock()
        mockAccountingDB._getConnection.return_value = {"OK": False, "Message": "No connection"}
        mockAccountingDB.retrieveBucketedData.return_value = {"OK": True, "Value": []}
        mockAccountingDB.calculateBucketLengthForTime.return_value = "BucketLength"
        obj = self.classsTested(mockAccountingDB, None)

        self.assertRaises(KeyError, obj._reportPhysicalSpace, {})
        self.assertRaises(IndexError, obj._reportPhysicalSpace, {"groupingFields": []})
        self.assertRaises(TypeError, obj._reportPhysicalSpace, {"groupingFields": [1, 2]})
        self.assertRaises(TypeError, obj._reportPhysicalSpace, {"groupingFields": [1, [2]]})
        self.assertRaises(TypeError, obj._reportPhysicalSpace, {"groupingFields": ["1", "2"]})
        self.assertRaises(KeyError, obj._reportPhysicalSpace, {"groupingFields": ["1", [2]]})
        self.assertRaises(KeyError, obj._reportPhysicalSpace, {"groupingFields": ["1", [2, 2]], "startTime": None})
        self.assertRaises(
            KeyError, obj._reportPhysicalSpace, {"groupingFields": ["1", [2, 2]], "startTime": None, "endTime": None}
        )
        self.assertRaises(
            TypeError,
            obj._reportPhysicalSpace,
            {"groupingFields": ["1", [2, 2]], "startTime": None, "endTime": None, "condDict": None},
        )

    def test_reportPhysicalFiles(self):
        """test the method "_reportPhysicalFiles"."""

        mockAccountingDB = mock.Mock()
        mockAccountingDB._getConnection.return_value = {"OK": False, "Message": "No connection"}
        mockAccountingDB.retrieveBucketedData.return_value = {"OK": True, "Value": []}
        mockAccountingDB.calculateBucketLengthForTime.return_value = "BucketLength"
        obj = self.classsTested(mockAccountingDB, None)

        self.assertRaises(KeyError, obj._reportPhysicalFiles, {})
        self.assertRaises(IndexError, obj._reportPhysicalFiles, {"groupingFields": []})
        self.assertRaises(TypeError, obj._reportPhysicalFiles, {"groupingFields": [1, 2]})
        self.assertRaises(TypeError, obj._reportPhysicalFiles, {"groupingFields": [1, [2]]})
        self.assertRaises(TypeError, obj._reportPhysicalFiles, {"groupingFields": ["1", "2"]})
        self.assertRaises(KeyError, obj._reportPhysicalFiles, {"groupingFields": ["1", [2]]})
        self.assertRaises(KeyError, obj._reportPhysicalFiles, {"groupingFields": ["1", [2, 2]], "startTime": None})
        self.assertRaises(
            KeyError, obj._reportPhysicalFiles, {"groupingFields": ["1", [2, 2]], "startTime": None, "endTime": None}
        )
        self.assertRaises(
            TypeError,
            obj._reportPhysicalFiles,
            {"groupingFields": ["1", [2, 2]], "startTime": None, "endTime": None, "condDict": None},
        )

    def test_plotCatalogSpace(self):
        """test the method "_plotCatalogSpace"."""

        obj = self.classsTested(None, None)
        self.assertRaises(TypeError, obj._plotCatalogSpace, None, None, None)
        self.assertRaises(KeyError, obj._plotCatalogSpace, {}, None, None)
        self.assertRaises(KeyError, obj._plotCatalogSpace, {"startTime": "startTime"}, None, None)
        self.assertRaises(
            TypeError, obj._plotCatalogSpace, {"startTime": "startTime", "endTime": "endTime"}, None, None
        )
        self.assertRaises(KeyError, obj._plotCatalogSpace, {"startTime": "startTime", "endTime": "endTime"}, {}, None)
        self.assertRaises(
            KeyError,
            obj._plotCatalogSpace,
            {"startTime": "startTime", "endTime": "endTime"},
            {"granularity": "granularity"},
            None,
        )
        self.assertRaises(
            KeyError,
            obj._plotCatalogSpace,
            {"startTime": "startTime", "endTime": "endTime", "grouping": "grouping"},
            {"granularity": "granularity"},
            None,
        )
        self.assertRaises(
            KeyError,
            obj._plotCatalogSpace,
            {"startTime": "startTime", "endTime": "endTime", "grouping": "grouping"},
            {"granularity": "granularity", "graphDataDict": "graphDataDict"},
            None,
        )

    def test_plotCatalogFiles(self):
        """test the method "_plotCatalogFiles"."""

        obj = self.classsTested(None, None)
        self.assertRaises(TypeError, obj._plotCatalogFiles, None, None, None)
        self.assertRaises(KeyError, obj._plotCatalogFiles, {}, None, None)
        self.assertRaises(KeyError, obj._plotCatalogFiles, {"startTime": "startTime"}, None, None)
        self.assertRaises(
            TypeError, obj._plotCatalogFiles, {"startTime": "startTime", "endTime": "endTime"}, None, None
        )
        self.assertRaises(KeyError, obj._plotCatalogFiles, {"startTime": "startTime", "endTime": "endTime"}, {}, None)
        self.assertRaises(
            KeyError,
            obj._plotCatalogFiles,
            {"startTime": "startTime", "endTime": "endTime"},
            {"granularity": "granularity"},
            None,
        )
        self.assertRaises(
            KeyError,
            obj._plotCatalogFiles,
            {"startTime": "startTime", "endTime": "endTime", "grouping": "grouping"},
            {"granularity": "granularity"},
            None,
        )
        self.assertRaises(
            KeyError,
            obj._plotCatalogFiles,
            {"startTime": "startTime", "endTime": "endTime", "grouping": "grouping"},
            {"granularity": "granularity", "graphDataDict": "graphDataDict"},
            None,
        )

    def test_plotPhysicalSpace(self):
        """test the method "_plotPhysicalSpace"."""

        obj = self.classsTested(None, None)
        self.assertRaises(TypeError, obj._plotPhysicalSpace, None, None, None)
        self.assertRaises(KeyError, obj._plotPhysicalSpace, {}, None, None)
        self.assertRaises(KeyError, obj._plotPhysicalSpace, {"startTime": "startTime"}, None, None)
        self.assertRaises(
            TypeError, obj._plotPhysicalSpace, {"startTime": "startTime", "endTime": "endTime"}, None, None
        )
        self.assertRaises(KeyError, obj._plotPhysicalSpace, {"startTime": "startTime", "endTime": "endTime"}, {}, None)
        self.assertRaises(
            KeyError,
            obj._plotPhysicalSpace,
            {"startTime": "startTime", "endTime": "endTime"},
            {"granularity": "granularity"},
            None,
        )
        self.assertRaises(
            KeyError,
            obj._plotPhysicalSpace,
            {"startTime": "startTime", "endTime": "endTime", "grouping": "grouping"},
            {"granularity": "granularity"},
            None,
        )
        self.assertRaises(
            KeyError,
            obj._plotPhysicalSpace,
            {"startTime": "startTime", "endTime": "endTime", "grouping": "grouping"},
            {"granularity": "granularity", "graphDataDict": "graphDataDict"},
            None,
        )

    def test_plotPhysicalFiles(self):
        """test the method "_plotPhysicalFiles"."""

        obj = self.classsTested(None, None)
        self.assertRaises(TypeError, obj._plotPhysicalFiles, None, None, None)
        self.assertRaises(KeyError, obj._plotPhysicalFiles, {}, None, None)
        self.assertRaises(KeyError, obj._plotPhysicalFiles, {"startTime": "startTime"}, None, None)
        self.assertRaises(
            TypeError, obj._plotPhysicalFiles, {"startTime": "startTime", "endTime": "endTime"}, None, None
        )
        self.assertRaises(KeyError, obj._plotPhysicalFiles, {"startTime": "startTime", "endTime": "endTime"}, {}, None)
        self.assertRaises(
            KeyError,
            obj._plotPhysicalFiles,
            {"startTime": "startTime", "endTime": "endTime"},
            {"granularity": "granularity"},
            None,
        )
        self.assertRaises(
            KeyError,
            obj._plotPhysicalFiles,
            {"startTime": "startTime", "endTime": "endTime", "grouping": "grouping"},
            {"granularity": "granularity"},
            None,
        )
        self.assertRaises(
            KeyError,
            obj._plotPhysicalFiles,
            {"startTime": "startTime", "endTime": "endTime", "grouping": "grouping"},
            {"granularity": "granularity", "graphDataDict": "graphDataDict"},
            None,
        )


################################################################################
#############################################################################


if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(DataStoragePlotterTestCase)
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(DataStoragePlotterTestCase))
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(DataStoragePlotterUnitTestCrashes))
    testResult = unittest.TextTestRunner(verbosity=2).run(suite)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""DIRAC Transformation DB.

Transformation database is used to collect and serve the necessary
information in order to automate the task of job preparation for high
level transformations. This class is typically used as a base class for
more specific data processing databases
"""
import threading
import copy
import re
from collections import defaultdict

from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.TransformationSystem.DB.TransformationDB import TransformationDB as DIRACTransformationDB
from DIRAC.Core.Utilities.List import intListToString, breakListIntoChunks


class TransformationDB(DIRACTransformationDB):
    """Extension of the DIRAC Transformation DB."""

    def __init__(self, dbname=None, dbconfig=None, dbIn=None, parentLogger=None):
        """The standard constructor takes the database name (dbname) and the name
        of the configuration section (dbconfig)"""
        DIRACTransformationDB.__init__(self, dbname, dbconfig, dbIn, parentLogger=parentLogger)
        self.lock = threading.Lock()
        self.queryFields = (
            "SimulationConditions",
            "DataTakingConditions",
            "ProcessingPass",
            "FileType",
            "EventType",
            "ConfigName",
            "ConfigVersion",
            "ProductionID",
            "DataQualityFlag",
            "ExtendedDQOK",
            "SMOG2",
            "StartRun",
            "EndRun",
            "EndDate",
            "Visible",
            "RunNumbers",
            "TCK",
        )
        self.intFields = ("EventType", "ProductionID", "StartRun", "EndRun")
        self.transRunParams = ("TransformationID", "RunNumber", "SelectedSite", "Status", "LastUpdate")
        self.allowedStatusForTasks = ("Unused", "ProbInFC")
        self.TRANSPARAMS.append("Hot")
        self.TRANSFILEPARAMS.extend(["RunNumber", "Size", "FileType", "RAWAncestors"])
        self.TASKSPARAMS.append("RunNumber")

    def deleteTransformation(self, transID, author="", connection=False):
        """Small extension to not forget to delete the BkQueries."""
        res = self.deleteBookkeepingQuery(transID, connection)
        if not res["OK"]:
            return res

        return DIRACTransformationDB.deleteTransformation(self, transID, author, connection)

    def cleanTransformation(self, transID, author="", connection=False):
        """Clean the transformation specified by name or id Extends DIRAC one for
        deleting the unused runs metadata."""
        res = self._getConnectionTransID(connection, transID)
        if not res["OK"]:
            return res
        connection = res["Value"]["Connection"]

        # deleting runs metadata
        req = f"SELECT DISTINCT RunNumber FROM TransformationRuns WHERE TransformationID = {transID}"
        res = self._query(req)
        if not res["OK"]:
            gLogger.error(f"Failure executing {str(req)}")
            return res
        runsMaybeToDelete = [r[0] for r in res["Value"]]

        req = f"SELECT DISTINCT RunNumber FROM TransformationRuns WHERE TransformationID != {transID}"
        res = self._query(req)
        if not res["OK"]:
            gLogger.error(f"Failure executing {str(req)}")
            return res
        runsToKeep = [r[0] for r in res["Value"]]

        runIDsToBeDeleted = list(set(runsMaybeToDelete) - set(runsToKeep))

        if runIDsToBeDeleted:
            res = self.deleteRunsMetadata(
                {"RunNumber": runIDsToBeDeleted, "Name": ["TCK", "CondDb", "DDDB"]}, connection
            )
            if not res["OK"]:
                return res

        res = DIRACTransformationDB.cleanTransformation(self, transID, connection=connection)
        if not res["OK"]:
            return res

        res = self.removeStoredJobDescription(transID, connection=connection)
        if not res["OK"]:
            return res

        res = self.__cleanTransformationRuns(transID, connection=connection)
        if not res["OK"]:
            return res
        return S_OK(transID)

    def getTasksForSubmission(
        self, transID, numTasks=1, site="", statusList=["Created"], older=None, newer=None, connection=False
    ):
        """extends base class including the run metadata."""
        tasksDict = DIRACTransformationDB.getTasksForSubmission(
            self, transID, numTasks, site, statusList, older, newer, connection
        )
        if not tasksDict["OK"]:
            return tasksDict

        tasksDict = tasksDict["Value"]
        runNumbers = []
        for taskForSumbission in tasksDict.values():
            run = taskForSumbission.get("RunNumber")
            if run and run not in runNumbers:
                runNumbers.append(run)

        if runNumbers:
            runsMetadata = self.getRunsMetadata(runNumbers, connection)
            if not runsMetadata["OK"]:
                return runsMetadata

            runsMetadata = runsMetadata["Value"]
            for taskForSumbission in tasksDict.values():
                try:
                    taskForSumbission["RunMetadata"] = runsMetadata[taskForSumbission["RunNumber"]]
                except KeyError:
                    continue

        return S_OK(tasksDict)

    def setHotFlag(self, transID, flag=False, connection=False):
        """Simply set the hot flag."""
        connection = self.__getConnection(connection)
        return self._update(
            "UPDATE Transformations SET Hot = %s WHERE TransformationID = %d" % (flag, int(transID)), conn=connection
        )

    #############################################################################
    #
    # Managing the BkQueries table

    def deleteBookkeepingQuery(self, transID, connection=False):
        """Delete the specified query from the database."""
        connection = self.__getConnection(connection)
        return self._update("DELETE FROM BkQueriesNew WHERE TransformationID=%d" % int(transID), conn=connection)

    def setBookkeepingQueryEndRun(self, transID, runNumber, connection=False):
        """Set the EndRun for the supplied transformation."""
        res = self._getConnectionTransID(connection, transID)
        if not res["OK"]:
            return res
        connection = res["Value"]["Connection"]
        res = self.getBookkeepingQuery(transID, connection=connection)
        if not res["OK"]:
            return res
        startRun = res["Value"].get("StartRun")
        endRun = res["Value"].get("EndRun")
        if endRun and runNumber < endRun:
            return S_ERROR("EndRun can not be reduced!")
        if startRun and startRun > runNumber:
            return S_ERROR("EndRun is before StartRun!")
        req = (
            "INSERT INTO BkQueriesNew(TransformationID,ParameterName,ParameterValue) VALUES(%d,'EndRun',%d) \
    ON DUPLICATE KEY UPDATE ParameterValue = %d"
            % (transID, runNumber, runNumber)
        )
        return self._update(req, conn=connection)

    def setBookkeepingQueryStartRun(self, transID, runNumber, connection=False):
        """Set the StartRun for the supplied transformation."""
        res = self._getConnectionTransID(connection, transID)
        if not res["OK"]:
            return res
        connection = res["Value"]["Connection"]
        res = self.getBookkeepingQuery(transID, connection=connection)
        if not res["OK"]:
            return res
        endRun = res["Value"].get("EndRun")
        startRun = res["Value"].get("StartRun")
        if startRun and runNumber > startRun:
            return S_ERROR("StartRun can not be increased!")
        if endRun and runNumber > endRun:
            return S_ERROR("StartRun cannot be after EndRun!")
        req = (
            "INSERT INTO BkQueriesNew(TransformationID,ParameterName,ParameterValue) VALUES(%d,'StartRun',%d) \
    ON DUPLICATE KEY UPDATE ParameterValue = %d"
            % (transID, runNumber, runNumber)
        )
        return self._update(req, conn=connection)

    def addBookkeepingQueryRunList(self, transID, runList, connection=False):
        """Adds the list of runs."""
        res = self._getConnectionTransID(connection, transID)
        if not res["OK"]:
            return S_ERROR("Failed to get Connection to TransformationDB")
        connection = res["Value"]["Connection"]
        res = self.getBookkeepingQuery(transID, connection=connection)
        if not res["OK"]:
            return S_ERROR("Cannot retrieve BkQuery")
        if "StartRun" in res["Value"] or "EndRun" in res["Value"]:
            return S_ERROR("RunList incompatible with start or end run")
        runsInQuery = set()
        try:
            runNumbers = res["Value"].get("RunNumbers", [])
            if runNumbers != "All":
                if not isinstance(runNumbers, list):
                    runNumbers = [runNumbers]
                runsInQuery = {int(run) for run in runNumbers}
            else:
                return S_ERROR("Cannot add runs to 'All'")
            runsInQuery |= {int(run) for run in runList}
        except ValueError as e:
            return S_ERROR(f"RunList invalid: {repr(e)}")
        if len(runsInQuery) > 999:
            return S_ERROR("RunList bigger the 1000 not allowed because of Oracle limitations!!!")
        value = ";;;".join(str(x) for x in sorted(runsInQuery))
        req = (
            "UPDATE BkQueriesNew SET ParameterValue='%s' WHERE TransformationID = %d AND ParameterName='RunNumbers'"
            % (value, transID)
        )
        self._update(req, conn=connection)
        return S_OK()

    def addBookkeepingQuery(self, transID, queryDict, connection=False):
        """Add a new Bookkeeping query specification."""
        connection = self.__getConnection(connection)
        values = []
        for field in self.queryFields:
            if field not in queryDict:
                if field in ["StartRun", "EndRun"]:
                    value = 0
                else:
                    value = "All"
            else:
                value = queryDict[field]

                if isinstance(value, (int,) + (float,)):
                    value = str(value)
                if isinstance(value, (list, tuple)):
                    value = ";;;".join(str(x) for x in value)
            values.append(value)

        # Insert the new bk query
        req = "INSERT INTO BkQueriesNew (TransformationID, ParameterName,ParameterValue) VALUES "
        for field, value in zip(self.queryFields, values):
            req += "(%d,'%s','%s'), " % (transID, field, value)
        req = req.strip().rstrip(",")

        res = self._update(req, conn=connection)
        if not res["OK"]:
            return res
        return S_OK(transID)

    def getTransformationsWithBkQueries(self, transIDs=None, connection=False):
        """Get those Transformations that have a BkQuery."""
        connection = self.__getConnection(connection)
        req = "SELECT DISTINCT TransformationID FROM BkQueriesNew"
        if transIDs:
            req = req + f" WHERE TransformationID IN ({', '.join(str(t) for t in transIDs)})"
        res = self._query(req, conn=connection)
        if res["OK"]:
            res = S_OK([tID[0] for tID in res["Value"]])
        return res

    def getBookkeepingQuery(self, transID, connection=False):
        """Get the bookkeeping query parameters."""
        result = self.getBookkeepingQueries([transID], connection)
        if result["OK"]:
            result["Value"] = result["Value"][transID]
        return result

    def getBookkeepingQueries(self, transIDs, connection=False):
        """Get the bookkeeping query parameters."""
        connection = self.__getConnection(connection)
        req = (
            "SELECT TransformationID, ParameterName, ParameterValue FROM BkQueriesNew "
            f"WHERE TransformationID IN ({', '.join(str(t) for t in transIDs)})"
        )
        res = self._query(req, conn=connection)
        if not res["OK"]:
            return res
        bkDict = defaultdict(dict)
        for row in res["Value"]:
            transID = row[0]
            parameter = row[1]
            value = row[2]
            if value and value != "All":
                if re.search(";;;", str(value)):
                    value = value.split(";;;")
                if parameter in self.intFields:
                    if isinstance(value, str):
                        value = int(value)
                    if isinstance(value, list):
                        value = [int(x) for x in value]
                    if not value:
                        continue
                bkDict[transID][parameter] = value
        for transID in transIDs:
            if transID not in bkDict:
                return S_ERROR("BkQuery %d not found" % int(transID))
        return S_OK(dict(bkDict))

    def __insertExistingTransformationFiles(self, transID, fileTuplesList, connection=False):
        """extends DIRAC.__insertExistingTransformationFiles Does not add userSE
        and adds runNumber."""

        gLogger.info(f"Inserting {len(fileTuplesList)} files in TransformationFiles")
        # splitting in various chunks, in case it is too big
        for fileTuples in breakListIntoChunks(fileTuplesList, 10000):
            gLogger.verbose(
                f"Adding first {len(fileTuples)} files in TransformationFiles (out of {len(fileTuplesList)})"
            )
            req = "INSERT INTO TransformationFiles (TransformationID,Status,TaskID,FileID, \
      TargetSE,LastUpdate,RunNumber,Size,FileType,RAWAncestors) VALUES"
            candidates = False

            for ft in fileTuples:
                (
                    _lfn,
                    originalID,
                    fileID,
                    status,
                    taskID,
                    targetSE,
                    _usedSE,
                    _errorCount,
                    _lastUpdate,
                    _insertTime,
                    runNumber,
                    size,
                    fileType,
                    rawAncestors,
                ) = ft[:14]
                if status not in ("Removed",):
                    candidates = True
                    if not re.search("-", status):
                        status = f"{status}-inherited"
                        if taskID:
                            taskID = 1000000 * int(originalID) + int(taskID)
                    req = "%s (%d,'%s',%s,%d,'%s',UTC_TIMESTAMP(),%s,%s,'%s',%s)," % (
                        req,
                        transID,
                        status,
                        taskID,
                        fileID,
                        targetSE,
                        runNumber,
                        size,
                        fileType,
                        rawAncestors,
                    )
            if not candidates:
                continue
            req = req.rstrip(",")
            res = self._update(req, conn=connection)
            if not res["OK"]:
                return res

        # We must also copy the run table entries if any
        result = self.getTransformationRuns({"TransformationID": originalID})
        if not result["OK"]:
            return result
        for runDict in res["Value"]:
            runID = runDict["RunNumber"]
            selectedSite = runDict["SelectedSite"]
            status = runDict["Status"]
            res = self.insertTransformationRun(
                transID, runID, selectedSite=selectedSite, status=status, connection=connection
            )
            if not res["OK"]:
                return res

        return S_OK()

    #############################################################################
    #
    # Managing the TransformationTasks table
    #

    def addTaskForTransformation(self, transID, lfns=[], se="Unknown", connection=False):
        """Create a new task with the supplied files for a transformation."""
        res = self._getConnectionTransID(connection, transID)
        if not res["OK"]:
            return res
        connection = res["Value"]["Connection"]
        # Be sure the all the supplied LFNs are known to the database for the supplied transformation
        fileIDs = []
        runID = 0
        if lfns:
            res = self.getTransformationFiles(
                condDict={"TransformationID": transID, "LFN": lfns}, connection=connection
            )
            if not res["OK"]:
                return res
            foundLfns = set()
            for fileDict in res["Value"]:
                fileIDs.append(fileDict["FileID"])
                runID = fileDict.get("RunNumber")
                if not runID:
                    runID = 0
                lfn = fileDict["LFN"]
                if fileDict["Status"] in self.allowedStatusForTasks:
                    foundLfns.add(lfn)
                else:
                    gLogger.error(
                        f"Supplied file not in {self.allowedStatusForTasks} status but {fileDict['Status']}", lfn
                    )
            unavailableLfns = set(lfns) - foundLfns
            if unavailableLfns:
                return S_ERROR("Not all supplied files available in the transformation database")

        # Insert the task into the jobs table and retrieve the taskID
        self.lock.acquire()
        req = (
            "INSERT INTO TransformationTasks(TransformationID, ExternalStatus, ExternalID, \
    TargetSE, CreationTime, LastUpdateTime, RunNumber) VALUES\
     (%d,'%s','%d','%s', UTC_TIMESTAMP(), UTC_TIMESTAMP(),%d);"
            % (transID, "Created", 0, se, runID)
        )
        res = self._update(req, conn=connection)
        if not res["OK"]:
            self.lock.release()
            gLogger.error("Failed to publish task for transformation", res["Message"])
            return res

        # TaskID is computed by a trigger, which sets the local variable @last (per connection)
        # @last is the last insert TaskID. With multi-row inserts, will be the first new TaskID inserted.
        # The trigger TaskID_Generator must be present with the InnoDB schema (defined in TransformationDB.sql)
        res = self._query("SELECT @last;", conn=connection)

        self.lock.release()
        if not res["OK"]:
            return res
        taskID = int(res["Value"][0][0])
        gLogger.verbose("Published task %d for transformation %d." % (taskID, transID))
        # If we have input data then update their status, and taskID in the transformation table
        if lfns:
            res = self.__insertTaskInputs(transID, taskID, lfns, connection=connection)
            if not res["OK"]:
                self.__removeTransformationTask(transID, taskID, connection=connection)
                return res
            res = self.__assignTransformationFile(transID, taskID, se, fileIDs, connection=connection)
            if not res["OK"]:
                self.__removeTransformationTask(transID, taskID, connection=connection)
                return res
        return S_OK(taskID)

    #############################################################################
    #
    # Managing the TransformationRuns table
    #

    def getTransformationRuns(
        self,
        condDict=None,
        older=None,
        newer=None,
        timeStamp="LastUpdate",
        orderAttribute=None,
        limit=None,
        connection=False,
    ):
        """Gets the transformation runs registered (usual query)"""

        connection = self.__getConnection(connection)
        selectDict = {}
        if condDict:
            for key, val in condDict.items():
                if key in self.transRunParams:
                    selectDict[key] = val
        req = "SELECT {} FROM TransformationRuns {}".format(
            intListToString(self.transRunParams),
            self.buildCondition(
                selectDict, older=older, newer=newer, timeStamp=timeStamp, orderAttribute=orderAttribute, limit=limit
            ),
        )
        res = self._query(req, conn=connection)
        if not res["OK"]:
            return res
        webList = []
        resultList = []
        for row in res["Value"]:
            # Prepare the structure for the web
            rList = []
            transDict = {}
            for param, item in zip(self.transRunParams, row):
                transDict[param] = item
                rList.append(item if isinstance(item, int) else str(item))
            webList.append(rList)
            resultList.append(transDict)
        result = S_OK(resultList)
        result["Records"] = webList
        result["ParameterNames"] = copy.copy(self.transRunParams)
        return result

    def getTransformationRunStats(self, transIDs, connection=False):
        """Gets counters of runs (taken from the TransformationFiles table)"""
        connection = self.__getConnection(connection)
        res = self.getCounters(
            "TransformationFiles", ["TransformationID", "RunNumber", "Status"], {"TransformationID": transIDs}
        )
        if not res["OK"]:
            return res
        transRunStatusDict = {}
        for attrDict, count in res["Value"]:
            transID = attrDict["TransformationID"]
            runID = attrDict["RunNumber"]
            status = attrDict["Status"]
            transRunStatusDict[transID][runID]["Total"] = (
                transRunStatusDict.setdefault(transID, {}).setdefault(runID, {}).setdefault("Total", 0) + count
            )
            transRunStatusDict[transID][runID][status] = count
        return S_OK(transRunStatusDict)

    def addTransformationRunFiles(self, transID, runID, lfns, connection=False):
        """Adds the RunID to the TransformationFiles table."""
        if not lfns:
            return S_ERROR("Zero length LFN list")
        res = self._getConnectionTransID(connection, transID)
        if not res["OK"]:
            return res
        lfnsDict = dict.fromkeys(lfns, {"RunNumber": runID})
        res = self.setParameterToTransformationFiles(transID, lfnsDict)
        if not res["OK"]:
            return res
        fileIDs = res["Value"]
        successful = dict.fromkeys(fileIDs.values(), "Added")
        failed = dict.fromkeys(set(lfns) - set(successful), "Missing")
        # Now update the TransformationRuns to include the newly updated files
        req = (
            "UPDATE TransformationRuns SET LastUpdate = UTC_TIMESTAMP() \
    WHERE TransformationID = %d and RunNumber = %d"
            % (transID, runID)
        )
        res = self._update(req, conn=connection)
        if not res["OK"]:
            gLogger.error("Failed to update TransformationRuns table with LastUpdate", res["Message"])
        elif not res["Value"]:
            self.insertTransformationRun(transID, runID, connection=connection)
        resDict = {"Successful": successful, "Failed": failed}
        return S_OK(resDict)

    def setParameterToTransformationFiles(self, transID, lfnsDict, connection=False):
        """Sets a parameter in the TransformationFiles table."""
        res = self._getConnectionTransID(connection, transID)
        if not res["OK"]:
            return res
        connection = res["Value"]["Connection"]
        res = self.__getFileIDsForLfns(list(lfnsDict), connection=connection)
        if not res["OK"]:
            return res
        fileIDs = res["Value"][0]
        rDict = {fileID: lfnsDict[lfn] for fileID, lfn in fileIDs.items()}
        for fID, param in rDict.items():
            values = ",".join(f"`{k}` = NULL" if v is None else f"`{k}` = '{v}'" for k, v in param.items())
            query = "UPDATE TransformationFiles SET %s WHERE TransformationID = %d AND FileID = %d"
            query %= (values, transID, fID)
            res = self._update(query, conn=connection)
            if not res["OK"]:
                gLogger.error("Failed to update TransformationFiles table", res["Message"])
                return res
        return S_OK(fileIDs)

    def setTransformationRunStatus(self, transID, runIDs, status, connection=False):
        """Sets a status in the TransformationRuns table."""
        if not runIDs:
            return S_OK()
        if not isinstance(runIDs, list):
            runIDs = [runIDs]
        res = self._getConnectionTransID(connection, transID)
        if not res["OK"]:
            return res
        connection = res["Value"]["Connection"]
        req = (
            "UPDATE TransformationRuns SET Status = '%s', LastUpdate = UTC_TIMESTAMP() \
    WHERE TransformationID = %d and RunNumber in (%s)"
            % (status, transID, intListToString(runIDs))
        )
        res = self._update(req, conn=connection)
        if not res["OK"]:
            gLogger.error("Failed to update TransformationRuns table with Status", res["Message"])
        return res

    def setTransformationRunsSite(self, transID, runID, selectedSite, connection=False):
        """Sets the site for Transformation Runs."""
        res = self._getConnectionTransID(connection, transID)
        if not res["OK"]:
            return res
        connection = res["Value"]["Connection"]
        req = (
            "UPDATE TransformationRuns SET SelectedSite = '%s', LastUpdate = UTC_TIMESTAMP() \
    WHERE TransformationID = %d and RunNumber = %d"
            % (selectedSite, transID, runID)
        )
        res = self._update(req, conn=connection)
        if not res["OK"]:
            gLogger.error("Failed to update TransformationRuns table with SelectedSite", res["Message"])
        elif not res["Value"]:
            res = self.insertTransformationRun(transID, runID, selectedSite=selectedSite, connection=connection)
        return res

    def insertTransformationRun(self, transID, runID, selectedSite="", status=None, connection=False):
        """Inserts a new Run for a specific transformation."""
        if status is None:
            status = "Active"
        if selectedSite:
            req = (
                "INSERT INTO TransformationRuns (TransformationID,RunNumber,SelectedSite,Status,LastUpdate) \
      VALUES (%d,%d,'%s','%s',UTC_TIMESTAMP())"
                % (transID, runID, selectedSite, status)
            )
        else:
            req = (
                "INSERT INTO TransformationRuns (TransformationID,RunNumber,Status,LastUpdate) \
      VALUES (%d,%d,'%s',UTC_TIMESTAMP())"
                % (transID, runID, status)
            )
        res = self._update(req, conn=connection)
        if not res["OK"]:
            gLogger.error("Failed to insert to TransformationRuns table", res["Message"])
        return res

    def __cleanTransformationRuns(self, transID, connection=False):
        req = f"DELETE  FROM TransformationRuns WHERE TransformationID = {transID}"
        res = self._update(req, conn=connection)
        if not res["OK"]:
            gLogger.error(f"Failure executing {str(req)}")
        return res

    #############################################################################
    #
    # Managing the RunsMetadata table
    #

    def setRunsMetadata(self, runID, metadataDict, connection=False):
        """Add the metadataDict to runID (if already present, does nothing)"""
        connection = self.__getConnection(connection)
        for name, value in metadataDict.items():
            res = self.__insertRunMetadata(runID, name, value, connection)
            if not res["OK"]:
                return res
        return S_OK()

    def updateRunsMetadata(self, runID, metadataDict, connection=False):
        """Add the metadataDict to runID (if already present, does nothing)"""
        connection = self.__getConnection(connection)
        for name, value in metadataDict.items():
            res = self.__updateRunMetadata(runID, name, value, connection)
            if not res["OK"]:
                return res
        return S_OK()

    def __insertRunMetadata(self, runID, name, value, connection):
        if isinstance(runID, str):
            runID = int(runID)
        req = "INSERT INTO RunsMetadata (RunNumber, Name, Value) VALUES(%d, '%s', '%s')" % (runID, name, value)
        res = self._update(req, conn=connection)
        if not res["OK"]:
            if "1062: Duplicate entry" in res["Message"]:
                gLogger.debug(f"Failed to insert to RunsMetadata table: {res['Message']}")
                return S_OK()
            else:
                gLogger.error("Failed to insert to RunsMetadata table", res["Message"])
                return res
        gLogger.info("Inserted %s %s of run %d to RunsMetadata table" % (name, value, runID))
        return S_OK()

    def __updateRunMetadata(self, runID, name, value, connection):
        if isinstance(runID, str):
            runID = int(runID)
        req = "UPDATE RunsMetadata SET Value = %s WHERE RunNumber = %d AND Name = '%s'" % (value, runID, name)
        res = self._update(req, conn=connection)
        if not res["OK"]:
            gLogger.error("Failed to update RunsMetadata table", res["Message"])
            return res
        gLogger.info("Updated %s %s of run %d in RunsMetadata table" % (name, value, runID))
        return S_OK()

    def getRunsMetadata(self, runIDs, connection=False):
        """get meta of a run.

        RunIDs can be a list.
        """
        connection = self.__getConnection(connection)
        try:
            if isinstance(runIDs, (list, dict, set, tuple)):
                runIDs = ",".join(str(int(x)) for x in runIDs)
            else:
                runIDs = str(int(runIDs))
        except ValueError as e:
            gLogger.exception("Invalid run number", lException=e)
            return S_ERROR("Invalid run number")
        req = f"SELECT RunNumber, Name, Value FROM RunsMetadata WHERE RunNumber IN ({runIDs})"
        res = self._query(req, conn=connection)
        if not res["OK"]:
            gLogger.error(f"Failure executing {str(req)}")
            return res
        else:
            dictOfNameValue = {}
            for runID, name, val in res["Value"]:
                dictOfNameValue.setdefault(runID, {}).update({name: val})
            return S_OK(dictOfNameValue)

    def deleteRunsMetadata(self, condDict=None, connection=False):
        """delete meta of a run.

        RunIDs can be a list.
        """
        connection = self.__getConnection(connection)
        req = f"DELETE FROM RunsMetadata {self.buildCondition(condDict)}"
        res = self._update(req, conn=connection)
        if not res["OK"]:
            gLogger.error(f"Failure executing {str(req)}")
            return res
        return S_OK()

    def getRunsInCache(self, condDict=None, connection=False):
        """get which runNumnber are cached."""
        connection = self.__getConnection(connection)
        req = f"SELECT DISTINCT RunNumber FROM RunsMetadata {self.buildCondition(condDict)}"
        res = self._query(req, conn=connection)
        if not res["OK"]:
            gLogger.error(f"Failure executing {str(req)}")
            return res
        else:
            return S_OK([run[0] for run in res["Value"]])

    #############################################################################
    #
    # Managing the RunDestination table
    #

    def getDestinationForRun(self, runIDs, connection=False):
        """get destination of a run or a list of runs."""
        connection = self.__getConnection(connection)
        req = f"SELECT * FROM RunDestination WHERE RunNumber IN ({', '.join(str(runID) for runID in runIDs)})"
        res = self._query(req, conn=connection)
        if not res["OK"]:
            gLogger.error(f"Failure executing {str(req)}")
            return res
        return S_OK(dict(res["Value"]) if res["Value"] else {})

    def setDestinationForRun(self, runID, destination, connection=False):
        """set destination of a run."""
        connection = self.__getConnection(connection)
        req = "INSERT INTO RunDestination (RunNumber, Destination) VALUES (%d, '%s')" % (runID, destination)
        res = self._query(req, conn=connection)
        if not res["OK"]:
            gLogger.error(f"Failure executing {str(req)}")
        return res

    #############################################################################
    #
    # Managing the StoredJobDescription table
    #

    def addStoredJobDescription(self, transformationID, jobDescription, connection=False):
        """store a job description for transformationID."""
        connection = self.__getConnection(connection)
        res = self._escapeString(jobDescription)
        if not res["OK"]:
            return S_ERROR("Failed to parse the transformation body")
        jobDescription = res["Value"]
        req = "INSERT INTO StoredJobDescription (TransformationID, JobDescription) VALUES (%d, %s)" % (
            transformationID,
            jobDescription,
        )
        res = self._query(req, conn=connection)
        if not res["OK"]:
            gLogger.error(f"Failure executing {str(req)}")
            return res
        else:
            return S_OK()

    def getStoredJobDescription(self, transformationID, connection=False):
        """get the job description for transformationID."""
        connection = self.__getConnection(connection)
        req = "SELECT * FROM StoredJobDescription WHERE TransformationID = %d" % transformationID
        res = self._query(req, conn=connection)
        if not res["OK"]:
            gLogger.error(f"Failure executing {str(req)}")
            return res
        else:
            return S_OK(res["Value"])

    def removeStoredJobDescription(self, transformationID, connection=False):
        """remove the job description for transformationID."""
        connection = self.__getConnection(connection)
        req = "DELETE FROM StoredJobDescription WHERE TransformationID = %d" % transformationID
        res = self._query(req, conn=connection)
        if not res["OK"]:
            gLogger.error(f"Failure executing {str(req)}")
            return res
        else:
            return S_OK()

    def getStoredJobDescriptionIDs(self, connection=False):
        """gets a list of all the stored job description transformationIDs."""
        connection = self.__getConnection(connection)
        req = "SELECT TransformationID FROM StoredJobDescription"
        res = self._query(req, conn=connection)
        if not res["OK"]:
            gLogger.error(f"Failure executing {str(req)}")
            return res
        else:
            return S_OK(res["Value"])



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

from DIRAC.Core.Base.Script import Script


@Script()
def main():
    Script.setUsageMessage("Gives an overview of the grid resources status")
    Script.parseCommandLine()

    from LHCbDIRAC.Interfaces.API.DiracLHCb import DiracLHCb

    DiracLHCb().gridWeather(printOutput=True)


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2022 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Functions for converting production objects to the legacy format

This module contains functions which convert the ``pydantic`` models in
:py:mod:`.Models` into dictionaries similar to those provided by the LHCb
WebApp applications.
"""
import json

from DIRAC.Core.Utilities.ReturnValues import returnValueOrRaise
from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
from LHCbDIRAC.ProductionManagementSystem.Utilities.Models import (
    DataProduction,
    ProductionBase,
    ProductionStep,
    SimulationProduction,
)


STEP_NAME_MAPPING = {
    "ApplicationName": "App",
    "ApplicationVersion": "Ver",
    "CONDDB": "CDb",
    "DDDB": "DDDb",
    "DQTag": "DQT",
    "ExtraPackages": "EP",
    "OptionFiles": "Opt",
    "OptionsFormat": "OptF",
    "ProcessingPass": "Pass",
    "StepId": "Step",
    "StepName": "Name",
    "SystemConfig": "SConf",
    "Usable": "Use",
    "Visible": "Vis",
    "isMulticore": "IsM",
    "mcTCK": "mcTCK",
}

PRODUCTION_DICT_KEYS = [
    "HasSubrequest",
    "NumberOfEvents",
    "RequestPDG",
    "Comments",
    "RetentionRate",
    "RequestAuthor",
    "Description",
    "RequestType",
    "EventType",
    "SimCondDetail",
    "RequestWG",
    "SimCondID",
    "IsModel",
    "FastSimulationType",
    "RequestPriority",
    "RealNumberOfEvents",
    "RequestState",
    "Inform",
    "ProPath",
    "RequestName",
    "SimCondition",
    "ProID",
    "Extra",
    "ProDetail",
    "MasterID",
    "RequestID",
    "ParentID",
]


def make_subprod_legacy_dict(sub_production, parent_id):
    """Make a webapp-like dictionary for defining a sub-production-request"""
    data = {k: None for k in PRODUCTION_DICT_KEYS}

    data["ParentID"] = parent_id
    data.update(sub_production)

    # Junk properties, included for consistency
    data["MasterID"] = data["ParentID"]
    data["HasSubrequest"] = 0
    data["IsModel"] = 0

    return data


def step_to_legacy_dict(i: int, step: ProductionStep):
    """Make a webapp-like dictionary that can be used to search for a step in the bookkeeping

    :param i: The index of the step in the production
    :param step: The step to convert into a dictionary
    """
    result = {
        "ApplicationName": step.application.name,
        "ApplicationVersion": step.application.version,
        "ExtraPackages": ";".join(sorted(f"{p.name}.{p.version}" for p in step.data_pkgs)),
        "ProcessingPass": step.processing_pass,
        "StepName": step.name,
        "isMulticore": "Y" if step.multicore else "N",
        # TODO: Really old steps have Visible = None?
        "Visible": "Y" if step.visible else "N",
        # TODO: Appears to have never been used
        # TODO: Required for running local tests
        "mcTCK": "",
        # 'DQTag': step.dbtags.DQ or '',
        # TODO: Get rid of RuntimeProjects from the database
    }
    if step.application.nightly is not None:
        result["ApplicationVersion"] = step.application.json()
    if step.id is not None:
        result["StepId"] = step.id
    if step.dbtags:
        result["CONDDB"] = step.dbtags.CondDB or ""
        result["DDDB"] = step.dbtags.DDDB or ""
        result["DQTag"] = step.dbtags.DQTag or ""
    elif i > 1:
        result["CONDDB"] = "fromPreviousStep"
        result["DDDB"] = "fromPreviousStep"
        result["DQTag"] = "fromPreviousStep"
    else:
        assert i > 0, "Step index should be 1-based"
        result["CONDDB"] = ""
        result["DDDB"] = ""
        result["DQTag"] = ""

    # if step.application.binary_tag is not None:
    result["SystemConfig"] = step.application.binary_tag or ""

    if isinstance(step.options, list):
        # Legacy style options
        result["OptionFiles"] = ";".join(step.options)
    else:
        result["OptionFiles"] = step.options.json()

    if step.options_format:
        result["OptionsFormat"] = step.options_format

    result["Usable"] = "Obsolete" if step.obsolete else ("Yes" if step.ready else "Not ready")

    return result


def step_to_step_manager_dict(i: int, step: ProductionStep):
    """Make a webapp-like dictionary for creating a step in the bookkeeping

    :param i: The index of the step in the production
    :param step: The step to convert into a dictionary
    """
    result = {"Step": step_to_legacy_dict(i, step)}

    if step.input:
        result["InputFileTypes"] = [{"FileType": x.type, "Visible": "Y" if x.visible else "N"} for x in step.input]
    result["OutputFileTypes"] = [{"FileType": x.type, "Visible": "Y" if x.visible else "N"} for x in step.output]
    return result


def _step_to_production_manager_dict(i: int, step: ProductionStep):
    """Make a webapp-like dictionary for defining a step in the metadata of a production request

    :param i: The index of the step in the production
    :param step: The step to convert into a dictionary
    """
    legacy_dict = step_to_legacy_dict(i, step)

    detail = {}
    if step.input:
        detail[f"p{i}IFT"] = ",".join(x.type for x in step.input)
    if step.output:
        detail[f"p{i}OFT"] = ",".join(x.type for x in step.output)
    for key, value in legacy_dict.items():
        if key not in STEP_NAME_MAPPING:
            raise NotImplementedError(f"Found unknown {key=}")

        full_key = f"p{i}{STEP_NAME_MAPPING[key]}"
        if value is None:
            if key in ["SConf"]:
                detail[full_key] = value
        else:
            detail[full_key] = str(value)

    # TODO: This should be generated dynamically rather than living in the database
    kwargs = dict(
        i=i,
        ift=detail.get(f"p{i}IFT", ""),
        oft=detail[f"p{i}OFT"],
        textRuntimeProjects="",
        SystemConfig=step.application.binary_tag or "",
        mcTCK="",
        OptionsFormat="",
        DQTag="",
        StepId="None",
    )
    kwargs.update({k: v or "" for k, v in legacy_dict.items()})
    detail[f"p{i}Html"] = (
        "<b>Step {i}</b> "
        "{StepName}({StepId}/{ProcessingPass}) : {ApplicationName}-{ApplicationVersion}<br/>"
        "System config: {SystemConfig} MC TCK: {mcTCK}<br/>"
        "Options: {OptionFiles} Options format: {OptionsFormat} "
        "Multicore: {isMulticore}<br/>"
        "DDDB: {DDDB} Condition DB: {CONDDB} DQTag: {DQTag}<br/>"
        "Extra: {ExtraPackages} "
        "Runtime projects: {textRuntimeProjects}<br/>"
        "Visible: {Visible} Usable:{Usable}<br/>"
        "Input file types: {ift} "
        "Output file types: {oft}<br/><br/>"
    ).format(**kwargs)

    p_all = f"{step.application.name}-{step.application.version}"

    p_dsc = None
    if step.application.name and step.processing_pass and step.visible:
        p_dsc = step.processing_pass

    return detail, p_all, p_dsc


def production_to_legacy_dict(prod: ProductionBase):
    """Make a webapp-like dictionary for creating a production request"""
    request = {
        "RequestName": prod.name,
        "RequestType": prod.type,
        "RequestAuthor": prod.author,
        "RequestPriority": prod.priority,
        "RequestState": prod.state.value,
        "RequestWG": prod.wg,
        "Comments": prod.comment,
        # MC options
        "IsModel": 0,
        "Extra": None,
        "HasSubrequest": 0,
        "NumberOfEvents": -1,
        "RealNumberOfEvents": -1,
        # Junk
        "Description": None,
        "ProID": None,
        "RequestPDG": None,
        "MasterID": None,
        "ParentID": None,
    }
    sub_productions = []

    if prod.id is not None:
        request["RequestID"] = prod.id

    if isinstance(prod, SimulationProduction):
        request["FastSimulationType"] = prod.fast_simulation_type
        request["RetentionRate"] = prod.retention_rate
        if prod.mc_config_version is not None:
            request["Extra"] = json.dumps({"mcConfigVersion": prod.mc_config_version})

        request.update(_lookup_simulation_condition(prod.sim_condition))
        sub_productions = [
            {
                "EventType": event_type.id,
                "NumberOfEvents": event_type.num_events,
                "RealNumberOfEvents": event_type.num_events,
            }
            for event_type in prod.event_types
        ]

        if len(sub_productions) > 1:
            request["EventType"] = None
            request["NumberOfEvents"] = None
            request["RealNumberOfEvents"] = 0
            request["HasSubrequest"] = 1
        else:
            request.update(sub_productions.pop(0))

    elif isinstance(prod, DataProduction):
        request["SimCondDetail"] = prod.input_dataset.conditions_dict.json()
        request["SimCondID"] = returnValueOrRaise(
            BookkeepingClient().getDataTakingConditionID(prod.input_dataset.conditions_description)
        )
        request["SimCondition"] = prod.input_dataset.conditions_description
        request["EventType"] = prod.input_dataset.event_type
    else:
        raise NotImplementedError(type(prod))

    request["Inform"] = ",".join(prod.inform)

    request["ProDetail"] = {"pAll": [], "pDsc": []}
    for i, step in enumerate(prod.steps, start=1):
        detail, p_all, p_dsc = _step_to_production_manager_dict(i, step)
        request["ProDetail"].update({k: v for k, v in detail.items() if v != ""})
        request["ProDetail"]["pAll"].append(p_all)
        if p_dsc is not None:
            request["ProDetail"]["pDsc"].append(p_dsc)

    request["ProDetail"]["pAll"] = ",".join(request["ProDetail"]["pAll"])
    if not isinstance(prod, SimulationProduction) or prod.override_processing_pass is None:
        request["ProDetail"]["pDsc"] = "/".join(request["ProDetail"]["pDsc"])
    else:
        request["ProDetail"]["pDsc"] = prod.override_processing_pass
    request["ProPath"] = request["ProDetail"]["pDsc"]
    if len(request["ProPath"]) >= 127:
        raise ValueError(f"ProPath is too long: {request['ProPath']!r}")
    request["ProDetail"] = json.dumps(request["ProDetail"])

    return request, sub_productions


def _lookup_simulation_condition(sim_condition: str):
    """Create a dictionary for the given simulation condition

    This method queries the bookkeeping to get full metadata of the corresponding condition.
    """
    query = {"SimDescription": sim_condition}
    conditions = retValToListOfDict(BookkeepingClient().getSimulationConditions(query))
    # getSimulationConditions queries for 'SimDescription like "%sim_condition%"'
    # so filter out any extra results
    conditions = [x for x in conditions if x["SimDescription"] == sim_condition]
    if len(conditions) == 0:
        raise NotImplementedError(
            f"{query} is not known, an expert should create it using dirac-bookkeeping-simulationconditions-insert"
        )
    elif len(conditions) != 1:
        raise NotImplementedError(conditions)
    conditions = conditions[0]
    simcond_detail = {
        "BeamEnergy": conditions["BeamEnergy"],
        "Generator": conditions["Generator"],
        "Luminosity": conditions["Luminosity"],
        "MagneticField": conditions["MagneticField"],
        "G4settings": conditions["G4settings"],
        "BeamCond": conditions["BeamCond"],
        "DetectorCond": conditions["DetectorCond"],
    }
    return {
        "SimCondDetail": json.dumps(simcond_detail),
        "SimCondID": conditions["SimId"],
        "SimCondition": conditions["SimDescription"],
    }


def find_step_id(i: int, step: ProductionStep) -> int | None:
    """Query the bookkeeping for a already defined step, returning its ID or None"""
    legacy_dict = step_to_legacy_dict(i, step)
    # Don't query on NULL keys as the bookkeeping gets confused
    query = {k: v for k, v in legacy_dict.items() if v}
    query.pop("Usable", None)
    matches = retValToListOfDict(BookkeepingClient().getAvailableSteps(query))

    matched = []
    for match in matches:
        # The ProcessingPass is treated as a prefix by getAvailableSteps so check it's an exact match
        if match["ProcessingPass"] != legacy_dict["ProcessingPass"]:
            continue
        input_types = [
            ProductionStep.FileType(type=x["FileType"], visible={"Y": True, "N": False}[x["Visible"]])
            for x in retValToListOfDict(BookkeepingClient().getStepInputFiles(match["StepId"]))
        ]
        if step.input != input_types:
            continue
        output_types = [
            ProductionStep.FileType(type=x["FileType"], visible={"Y": True, "N": False}[x["Visible"]])
            for x in retValToListOfDict(BookkeepingClient().getStepOutputFiles(match["StepId"]))
        ]
        if step.output != output_types:
            continue
        match.pop("RuntimeProjects", None)
        break
    else:
        return None

    # Sanity check
    for key in set(legacy_dict) | set(match):
        if key in ["StepId", "Usable"]:
            continue

        actual_value = match[key]
        if actual_value == "NULL":
            actual_value = ""
        expected_value = legacy_dict.get(key, None)

        if actual_value:
            if actual_value == expected_value:
                continue
        else:
            if not actual_value and not expected_value:
                continue
        raise NotImplementedError(key, actual_value, expected_value, matched)

    return match["StepId"]


def retValToListOfDict(retVal) -> list[dict]:
    """Convert a Records-style bookkeeping response to a list of dictionaries"""
    retVal = returnValueOrRaise(retVal)
    return [dict(zip(retVal["ParameterNames"], record)) for record in retVal["Records"]]


def runs_to_input_query(runs):
    """Convert a list of runs to a query for the bookkeeping system

    The list of runs can contain any combination of:
    - single run numbers as a string or integer
    - run ranges as a string in the form "start:end" (inclusive)
    """
    for run in runs:
        if isinstance(run, int):
            run = str(run)
        if ":" in run:
            start, end = map(int, run.split(":"))
            yield from map(str, range(start, end + 1))
        else:
            yield run


def configure_input(pr, data, *, runs=None, startRun=None, endRun=None):
    """Configure the input for a ProductionRequest object."""
    scd = json.loads(data["SimCondDetail"])
    import ast

    if "[" in scd["inDataQualityFlag"]:
        scd["inDataQualityFlag"] = ",".join(ast.literal_eval(scd["inDataQualityFlag"]))

    pr.startRun = startRun or 0
    pr.endRun = endRun or 0
    if runs:
        pr.runsList = ",".join(runs_to_input_query(runs))

    pr.configName = scd["configName"]
    pr.configVersion = scd["configVersion"]
    pr.dqFlag = scd["inDataQualityFlag"].replace(" ", "")
    if scd["inExtendedDQOK"]:
        pr.extendedDQOK = ",".join(scd["inExtendedDQOK"])
    if scd["inSMOG2State"]:
        pr.smog2State = scd["inSMOG2State"]
    pr.dataTakingConditions = data["SimCondition"]
    pr.processingPass = scd["inProPass"]
    pr.eventType = data["EventType"]
    if scd["inProductionID"] == "ALL":
        pr.bkFileType = scd["inFileType"]
        pr.bkQueries = ["Full"]
    else:
        pr.previousProdID = scd["inProductionID"]
        pr.bkQueries = ["fromPreviousProd"]
        pr.bkFileType = scd["inFileType"]



###############################################################################
# (c) Copyright 2024 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
import pytest
from copy import deepcopy

from DIRAC import S_OK

from LHCbDIRAC.BookkeepingSystem.Client.BKQuery import BKQuery, makeBKPath

AVAILABLE_FILE_TYPES = {
    "ParameterNames": ["FileType", "Description"],
    "Records": [
        ["DST", "A DST file"],
        ["MDST", "A smaller file"],
        ["ALLSTREAMS.ANAPROD_BSTOJPSIPHI.DST", None],
        ["HLT2.DST", "HLT2 DST file"],
    ],
}
AVAILABLE_FILE_TYPES["TotalRecords"] = len(AVAILABLE_FILE_TYPES["Records"])


@pytest.fixture
def fake_available_file_types(monkeypatch):
    monkeypatch.setattr(
        "LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient.BookkeepingClient.getAvailableFileTypes",
        lambda self: S_OK(deepcopy(AVAILABLE_FILE_TYPES)),
    )


def test_reversible(fake_available_file_types):
    query_dict = {
        "ConfigName": "MC",
        "ConfigVersion": "Dev",
        "ConditionDescription": "Beam6800GeV-expected-2024-MagUp-Nu7.6-25ns-Pythia8",
        "ProcessingPass": "/Sim10c/AnaProd-v0r0p6844333-MC_BuToD0K_D0ToHHHH_sig_Up_hlt1/2",
        "EventType": "12165042",
        "FileType": "HLT2.DST",
        "SimulationConditions": "Beam6800GeV-expected-2024-MagUp-Nu7.6-25ns-Pythia8",
    }
    query_path = makeBKPath(query_dict)
    assert (
        query_path == "/MC/Dev/Beam6800GeV-expected-2024-MagUp-Nu7.6-25ns-Pythia8/Sim10c/"
        "AnaProd-v0r0p6844333-MC_BuToD0K_D0ToHHHH_sig_Up_hlt1/2/12165042/HLT2.DST"
    )
    assert BKQuery().buildBKQuery(query_path) == query_dict | {"Visible": "Yes"}


def test_complex(fake_available_file_types):
    query_dict = BKQuery().buildBKQuery(
        "/MC/Dev/Beam6800GeV-expected-2024-MagUp-Nu7.6-25ns-Pythia8/Sim10c/"
        "AnaProd-v0r0p6844333-MC_BuToD0K_D0ToHHHH_sig_Up_hlt1/2/12165042/HLT2.DST",
        prods=[212760],
        fileTypes=["HLT2.DST"],
        visible=True,
    )
    assert query_dict == {
        "ConfigName": "MC",
        "ConfigVersion": "Dev",
        "ConditionDescription": "Beam6800GeV-expected-2024-MagUp-Nu7.6-25ns-Pythia8",
        "ProcessingPass": "/Sim10c/AnaProd-v0r0p6844333-MC_BuToD0K_D0ToHHHH_sig_Up_hlt1/2",
        "EventType": "12165042",
        "Production": 212760,
        "FileType": "HLT2.DST",
        "Visible": "Yes",
        "SimulationConditions": "Beam6800GeV-expected-2024-MagUp-Nu7.6-25ns-Pythia8",
    }



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""List simulation conditions from the Bookkeeping."""
import DIRAC
from DIRAC.Core.Base.Script import Script


@Script()
def main():
    Script.setUsageMessage(__doc__ + "\n".join(["Usage:", f"  {Script.scriptName} [option|cfgfile] ..."]))
    Script.parseCommandLine(ignoreErrors=True)

    from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

    exitCode = 0

    res = BookkeepingClient().getSimConditions()
    if res["OK"]:
        dbresult = res["Value"]
        for record in dbresult:
            print("SimId: " + str(record[0]).ljust(10))
            print("  SimDescription: " + str(record[1]).ljust(10))
            print("  BeamCond: " + str(record[2]).ljust(10))
            print("  BeamEnergy: " + str(record[3]).ljust(10))
            print("  Generator: " + str(record[4]).ljust(10))
            print("  MagneticField: " + str(record[5]).ljust(10))
            print("  DetectorCond: " + str(record[6]).ljust(10))
            print("  Luminosity: " + str(record[7]).ljust(10))
            print("  G4settings: " + str(record[8]).ljust(10))

    else:
        print("ERROR:", res["Message"])
        exitCode = 1

    DIRAC.exit(exitCode)


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
""" JSON encoder - slighltly modified version for LHCbDIRAC purposes
"""
import json
import datetime


class JSONDateTimeEncoder(json.JSONEncoder):
    """encoder of datetime objects."""

    def default(self, obj):  # pylint: disable=method-hidden
        if isinstance(obj, datetime.datetime):
            return {"dt": [obj.year, obj.month, obj.day, obj.hour, obj.minute, obj.second]}
        elif isinstance(obj, datetime.date):
            return {"d": [obj.year, obj.month, obj.day]}
        elif isinstance(obj, datetime.time):
            return {"t": [obj.hour, obj.minute, obj.second]}

        return super().default(obj)


class JSONDateTimeDecoder(json.JSONDecoder):
    """decoder of datetime objects."""

    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)

    def object_hook(self, obj):  # pylint: disable=method-hidden
        if "dt" in obj:
            obj = datetime.datetime(obj["dt"][0], obj["dt"][1], obj["dt"][2], obj["dt"][3], obj["dt"][4], obj["dt"][5])
        elif "d" in obj:
            obj = datetime.date(obj["d"][0], obj["d"][1], obj["d"][2])
        elif "t" in obj:
            obj = datetime.time(obj["t"][0], obj["t"][1], obj["t"][2])

        return obj


def dumps(obj):
    return json.dumps(obj, cls=JSONDateTimeEncoder)


def loads(obj):
    return json.loads(obj, cls=JSONDateTimeDecoder)


def load(fd):
    return json.load(fd, cls=JSONDateTimeDecoder)
